from functools import lru_cache
from typing import Tuple, List

import cv2 as cv
import numpy as np
import pandas as pd
from scipy.ndimage import binary_fill_holes, maximum_filter
from skimage import morphology as morph, img_as_float
from skimage.measure import regionprops, label, regionprops_table
from skimage.morphology import flood
from skimage import segmentation as seg
from utilitys.processing import *
from utilitys import fns

from s3a.constants import REQD_TBL_FIELDS as RTF, PRJ_ENUMS
from s3a.generalutils import cornersToFullBoundary, getCroppedImg, imgCornerVertices, \
  showMaskDiff, MaxSizeDict
from s3a.processing.processing import ImageProcess, GlobalPredictionProcess
from s3a.structures import BlackWhiteImg, XYVertices, ComplexXYVertices, NChanImg
from s3a.structures import GrayImg, RgbImg

def growSeedpoint(img: NChanImg, seeds: XYVertices, thresh: float) -> BlackWhiteImg:
  shape = np.array(img.shape[0:2])
  bwOut = np.zeros(shape, dtype=bool)
  # Turn x-y vertices into row-col seeds
  seeds = seeds[:, ::-1]
  # Remove seeds that don't fit in the image
  seeds = seeds[np.all(seeds >= 0, 1)]
  seeds = seeds[np.all(seeds < shape, 1)]

  for seed in seeds:
    for chan in range(img.shape[2]):
      curBwMask = flood(img[...,chan], tuple(seed), tolerance=thresh)
      bwOut |= curBwMask
  return bwOut

def colorLabelsWithMean(labelImg: GrayImg, refImg: NChanImg) -> RgbImg:
  outImg = np.empty(refImg.shape)
  labels = np.unique(labelImg)
  for curLabel in labels:
    curmask = labelImg == curLabel
    outImg[curmask,:] = refImg[curmask,:].reshape(-1,3).mean(0)
  return outImg

def growSeedpoint_cv_fastButErratic(img: NChanImg, seeds: XYVertices, thresh: float):
  if len(seeds) == 0:
    return np.zeros(img.shape[:2], bool)
  if img.ndim > 2:
    nChans = img.shape[2]
  else:
    nChans = 1
  thresh = int(np.clip(thresh, 0, 255))
  imRCShape = np.array(img.shape[:2])
  bwOut = np.zeros(imRCShape+2, 'uint8')
  # Throw away seeds outside image boundaries
  seeds = seeds[np.all(seeds < imRCShape, axis=1)]
  seeds = np.fliplr(seeds)
  mask = np.zeros(imRCShape+2, 'uint8')
  for seed in seeds:
    mask.fill(0)
    flooded = img.copy()
    seed = tuple(seed.flatten())
    _, _, curOut, _ = cv.floodFill(flooded, mask, seed, 255, (thresh,)*nChans, (thresh,)*nChans,8)
    bwOut |= curOut
  bwOut = bwOut[1:-1,1:-1]
  return bwOut.astype(bool)

def area_coord_regionTbl(_image: NChanImg):
  if not np.any(_image):
    return pd.DataFrame({'coords': [np.array([[]])], 'area': [0]})
  regionDict = regionprops_table(label(_image), properties=('coords', 'area'))
  _regionPropTbl = pd.DataFrame(regionDict)
  return _regionPropTbl


_historyMaskHolder = [np.array([[]], 'uint8')]

"""
0 = unspecified, 1 = background, 2 = foreground. Place inside list so reassignment
doesn't destroy object reference
"""
def format_vertices(image: NChanImg, fgVerts: XYVertices, bgVerts: XYVertices,
                    prevCompMask: BlackWhiteImg, firstRun: bool,
                    keepVertHistory=True):
  global _historyMaskHolder

  if firstRun or not keepVertHistory:
    _historyMask = np.zeros(image.shape[:2], 'uint8')
  else:
    _historyMask = _historyMaskHolder[0]

  asForeground = True
  # 0 = unspecified, 1 = background, 2 = foreground
  for fillClr, verts in enumerate([bgVerts, fgVerts], 1):
    if not verts.empty:
      cv.fillPoly(_historyMask, [verts], fillClr)

  if fgVerts.empty and bgVerts.empty:
    # Give whole image as input
    fgVerts = imgCornerVertices(image)
    fgVerts = cornersToFullBoundary(fgVerts)
    _historyMask[fgVerts.rows, fgVerts.cols] = 1
  _historyMaskHolder[0] = _historyMask
  curHistory = _historyMask.copy()
  if fgVerts.empty:
    # Invert the mask and paint foreground pixels
    asForeground = False
    # Invert the history mask too
    curHistory[_historyMask == 2] = 1
    curHistory[_historyMask == 1] = 2
    fgVerts = bgVerts
    bgVerts = XYVertices()

  if asForeground:
    foregroundAdjustedCompMask = prevCompMask.copy()
  else:
    foregroundAdjustedCompMask = ~prevCompMask

  # Default to bound slices that encompass the whole image
  bounds = np.array([[0, 0], image.shape[:2][::-1]])
  boundSlices = slice(*bounds[:,1]), slice(*bounds[:,0])
  return ProcessIO(image=image, summaryInfo=None, fgVerts=fgVerts, bgVerts=bgVerts, asForeground=asForeground,
                   historyMask=curHistory, prevCompMask=foregroundAdjustedCompMask,
                   origCompMask=prevCompMask, boundSlices=boundSlices)

def crop_to_local_area(image: NChanImg,
                       fgVerts: XYVertices,
                       bgVerts: XYVertices,
                       prevCompMask: BlackWhiteImg,
                       prevCompVerts: ComplexXYVertices,
                       viewbox: XYVertices,
                       historyMask: GrayImg,
                       reference='viewbox',
                       margin_pct=10
                       ):
  """
  :param reference:
    pType: list
    limits:
      - image
      - component
      - viewbox
      - roi
  """
  roiVerts = np.vstack([fgVerts, bgVerts])
  compVerts = np.vstack([prevCompVerts.stack(), roiVerts])
  if reference == 'image':
    allVerts = np.array([[0, 0], image.shape[:2]])
  elif reference == 'roi' and len(roiVerts) > 1:
    allVerts = roiVerts
  elif reference == 'component' and len(compVerts) > 1:
    allVerts = compVerts
  else:
    # viewbox or badly sized previous region/roi
    allVerts = np.vstack([viewbox, roiVerts])
  # Lots of points, use their bounded area
  try:
    vertArea_rowCol = (allVerts.max(0)-allVerts.min(0))[::-1]
  except ValueError:
    # 0-sized
    vertArea_rowCol = 0
  margin = int(round(max(vertArea_rowCol) * (margin_pct / 100)))
  cropped, bounds = getCroppedImg(image, allVerts, margin)
  vertOffset = bounds.min(0)
  for vertList in fgVerts, bgVerts:
    vertList -= vertOffset
    np.clip(vertList, a_min=[0,0], a_max=bounds[1,:]-1, out=vertList)
  boundSlices = slice(*bounds[:,1]), slice(*bounds[:,0])
  croppedCompMask = prevCompMask[boundSlices]
  curHistory = historyMask[boundSlices]

  rectThickness = int(max(1, *image.shape)*0.005)
  toPlot = cv.rectangle(image.copy(), tuple(bounds[0,:]), tuple(bounds[1,:]),
                        (255,0,0), rectThickness)
  info = {'name': 'Selected Area', 'image': toPlot}
  return ProcessIO(image=cropped, fgVerts=fgVerts, bgVerts=bgVerts, prevCompMask=croppedCompMask,
                   boundSlices=boundSlices, historyMask=curHistory, summaryInfo=info)

def apply_process_result(image: NChanImg, asForeground: bool,
                         prevCompMask: BlackWhiteImg, origCompMask: BlackWhiteImg,
                         boundSlices: Tuple[slice,slice]):
  if asForeground:
    bitOperation = np.bitwise_or
  else:
    # Add to background
    bitOperation = lambda curRegion, other: ~(curRegion | other)
  # The other basic operations need the rest of the component mask to work properly,
  # so expand the current area of interest only as much as needed. Returning to full size
  # now would incur unnecessary addtional processing times for the full-sized image
  outMask = origCompMask.copy()
  change = bitOperation(prevCompMask, image)
  outMask[boundSlices] = change
  foregroundPixs = np.c_[np.nonzero(outMask)]
  # Keep algorithm from failing when no foreground pixels exist
  if len(foregroundPixs) == 0:
    mins = [0,0]
    maxs = [1,1]
  else:
    mins = foregroundPixs.min(0)
    maxs = foregroundPixs.max(0)

  # Add 1 to max slice so stopping value is last foreground pixel
  newSlices = (slice(mins[0], maxs[0]+1), slice(mins[1], maxs[1]+1))
  return ProcessIO(image=outMask[newSlices], boundSlices=newSlices)

def return_to_full_size(image: NChanImg, origCompMask: BlackWhiteImg,
                        boundSlices: Tuple[slice]):
  out = np.zeros_like(origCompMask)
  if image.ndim > 2:
    image = image.mean(2).astype(int)
  out[boundSlices] = image

  infoMask = showMaskDiff(origCompMask[boundSlices], image)

  return ProcessIO(image=out, summaryInfo={'image': infoMask, 'name': 'Finalize Region'})

def fill_holes(image: NChanImg):
  return ProcessIO(image=binary_fill_holes(image))

def disallow_paint_tool(_image: NChanImg, fgVerts: XYVertices, bgVerts: XYVertices):
  if len(np.vstack([fgVerts, bgVerts])) < 2:
    raise ValueError('This algorithm requires an enclosed area to work.'
                              ' Only one vertex was given as an input.')
  return ProcessIO(image=_image)

def openClose():
  proc = ImageProcess('Open -> Close')
  def perform_op(image: NChanImg, radius=1, shape='rectangle'):
    """
    :param radius: Radius of the structuring element. Note that the total side length
      of the structuring element will be (2*radius)+1.
    :param shape:
      helpText: Structuring element shape
      pType: list
      limits:
        - rectangle
        - disk
        - diamond
    """
    if image.ndim > 2:
      image = image.mean(2)
    image = image.astype('uint8')
    ksize = [radius]
    if shape == 'rectangle':
      ksize = [ksize[0]*2+1]*2
    strel = getattr(morph, shape)(*ksize)
    outImg = cv.morphologyEx(image.copy(), cv.MORPH_OPEN, strel)
    outImg = cv.morphologyEx(outImg, cv.MORPH_CLOSE, strel)
    return outImg
  proc.addFunction(perform_op, name='Open -> Close', needsWrap=True)
  return proc

def keep_largest_comp(image: NChanImg):
  regionPropTbl = area_coord_regionTbl(image)
  out = np.zeros(image.shape, bool)
  coords = regionPropTbl.coords[regionPropTbl.area.argmax()]
  if coords.size == 0:
    return ProcessIO(image=out)
  out[coords[:,0], coords[:,1]] = True
  return ProcessIO(image=out)

def rm_small_comps(image: NChanImg, minSzThreshold=30):
  regionPropTbl = area_coord_regionTbl(image)
  validCoords = regionPropTbl.coords[regionPropTbl.area >= minSzThreshold]
  out = np.zeros(image.shape, bool)
  if len(validCoords) == 0:
    return ProcessIO(image=out)
  coords = np.vstack(validCoords)
  out[coords[:,0], coords[:,1]] = True
  return ProcessIO(image=out)

def basic_shapes(image: NChanImg, fgVerts: XYVertices, penSize=1, penShape='circle'):
  """
  Draws basic shapes with minimal pre- or post-processing.

  :param penSize: Size of the drawing pen
  :param penShape:
    helpText: Shape of the drawing pen
    pType: list
    limits:
      - circle
      - rectangle
  """
  out = np.zeros(image.shape[:2], dtype='uint8')
  drawFns = {
    'circle': lambda pt: cv.circle(out, tuple(pt), penSize//2, 1, -1),
    'rectangle': lambda pt: cv.rectangle(out, tuple(pt-penSize//2), tuple(pt+penSize//2), 1, -1)
  }
  try:
    drawFn = drawFns[penShape]
  except KeyError:
    raise ValueError(f"Can't understand shape {penShape}. Must be one of:\n"
                              f"{','.join(drawFns)}")
  if len(fgVerts) > 1 and penSize > 1:
    ComplexXYVertices([fgVerts]).toMask(out, 1, False, warnIfTooSmall=False)
  else:
    for vert in fgVerts:
      drawFn(vert)
  return out > 0

def convert_to_squares(image: NChanImg):
  outMask = np.zeros(image.shape, dtype=bool)
  for region in regionprops(label(image)):
    outMask[region.bbox[0]:region.bbox[2],region.bbox[1]:region.bbox[3]] = True
  return ProcessIO(image=outMask)

def basicOpsCombo():
  proc = ImageProcess('Basic Region Operations')
  toAdd: List[ImageProcess] = []
  for func in fill_holes, keep_largest_comp, rm_small_comps:
    toAdd.append(ImageProcess.fromFunction(func))
  proc.addProcess(toAdd[0])
  proc.addProcess(openClose())
  proc.addProcess(toAdd[1])
  proc.addProcess(toAdd[2])
  return proc


def _grabcutResultToMask(gcResult):
  return np.where((gcResult==2)|(gcResult==0), False, True)

def cv_grabcut(image: NChanImg, prevCompMask: BlackWhiteImg, fgVerts: XYVertices,
               noPrevMask: bool, historyMask: GrayImg, iters=5):
  if image.size == 0:
    return ProcessIO(image=np.zeros_like(prevCompMask))
  img = cv.cvtColor(image, cv.COLOR_RGB2BGR)
  # Turn foreground into x-y-width-height
  bgdModel = np.zeros((1,65),np.float64)
  fgdModel = np.zeros((1,65),np.float64)
  historyMask = historyMask.copy()
  historyMask[fgVerts.rows, fgVerts.cols] = 2

  mask = np.zeros(prevCompMask.shape, dtype='uint8')
  mask[prevCompMask == 1] = cv.GC_PR_FGD
  mask[prevCompMask == 0] = cv.GC_PR_BGD
  mask[historyMask == 2] = cv.GC_FGD
  mask[historyMask == 1] = cv.GC_BGD

  cvRect = np.array([fgVerts.min(0), fgVerts.max(0) - fgVerts.min(0)]).flatten()

  if noPrevMask:
    if cvRect[2] == 0 or cvRect[3] == 0:
      return ProcessIO(image=np.zeros_like(prevCompMask))
    mode = cv.GC_INIT_WITH_RECT
  else:
    mode = cv.GC_INIT_WITH_MASK
  cv.grabCut(img, mask, cvRect, bgdModel, fgdModel, iters, mode=mode)
  outMask = np.where((mask==2)|(mask==0), False, True)
  return ProcessIO(labels=outMask)

_qsInCache = MaxSizeDict(maxsize=5)
_qsOutCache = MaxSizeDict(maxsize=5)

def quickshift_seg(image: NChanImg, max_dist=10., kernel_size=5,
                   sigma=0.0):
  global _qsInCache, _qsOutCache
  # For max_dist of 0, the input isn't changed and it takes a long time
  key = (max_dist, kernel_size, sigma)
  if max_dist == 0:
    # Make sure output is still 1-channel
    if image.ndim > 2:
      segImg = image.mean(2).astype(int)
    else:
      segImg = image
  else:
    if image.ndim < 3:
      image = np.tile(image[:,:,None], (1,1,3))
    # First check if this image was the same as last time
    if np.array_equal(_qsInCache.get(key, None), image) and key in _qsOutCache:
      segImg =  _qsOutCache[key]
    else:
      segImg = seg.quickshift(image, kernel_size=kernel_size, max_dist=max_dist, sigma=sigma)
  _qsInCache[key] = image
  _qsOutCache[key] = segImg.copy()
  return ProcessIO(labels=segImg)
quickshift_seg.__doc__ = seg.quickshift.__doc__

# Taken from example page: https://scikit-image.org/docs/dev/auto_examples/segmentation/plot_morphsnakes.html
def morph_acwe(image: NChanImg, initialCheckerSize=6, iters=35, smoothing=3):
  image = img_as_float(image)
  if image.ndim > 2:
    image = image.mean(2)
  initLs = seg.checkerboard_level_set(image.shape, initialCheckerSize)
  outLevels = seg.morphological_chan_vese(image, iters, init_level_set=initLs, smoothing=smoothing)
  return ProcessIO(labels=outLevels)

def k_means(image: NChanImg, kVal=5, attempts=10):
  # Logic taken from https://docs.opencv.org/master/d1/d5c/tutorial_py_kmeans_opencv.html
  numChannels = 1 if image.ndim < 3 else image.shape[2]
  clrs = image.reshape(-1, numChannels)
  clrs = clrs.astype('float32')
  criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 10, 1.0)
  ret, lbls, imgMeans = cv.kmeans(clrs, kVal, None, criteria, attempts,
                             cv.KMEANS_RANDOM_CENTERS)
  # Now convert back into uint8, and make original image
  imgMeans = imgMeans.astype('uint8')
  lbls = lbls.reshape(image.shape[:2])

  return ProcessIO(labels=lbls, means=imgMeans)

def labelBoundaries_cv(labels: np.ndarray, thickness: int):
  """Code stolen and reinterpreted for cv from skimage.segmentation.boundaries"""
  if thickness % 2 == 0:
    thickness += 1
  thickness = max(thickness, 3)
  if labels.dtype not in [np.uint8, np.uint16, np.int16, np.float16, np.float32]:
    labels = labels.astype(np.uint16)
  strel = cv.getStructuringElement(cv.MORPH_RECT, (thickness, thickness))
  return cv.morphologyEx(labels, cv.MORPH_DILATE, strel) \
               != cv.morphologyEx(labels, cv.MORPH_ERODE, strel)

def binarize_labels(image: NChanImg, labels: BlackWhiteImg, fgVerts: XYVertices,
                    historyMask: BlackWhiteImg, touchingRoiOnly=True, useMeanColor=True,
                    lineThickness=2):
  """
  For a given binary image input, only keeps connected components that are directly in
  contact with at least one of the specified vertices. In essence, this function can make
  a wide variety of operations behave similarly to region growing.

  :param touchingRoiOnly: Whether to only keep labeled regions that are in contact
    with the current ROI
  :param useMeanColor: Whether to color the summary info image with mean values or
    (if *False*) just draw the boundaries around each label.
  :param lineThickness:
    helpText: How thick to draw label boundary and ROI vertices lines
  """
  if labels.ndim > 2:
    raise ValueError('Cannot handle multichannel labels.\n'
                     f'(labelss.shape={labels.shape})')
  seeds = cornersToFullBoundary(fgVerts, 50e3)[:, ::-1]
  seeds = np.clip(seeds, 0, np.array(labels.shape)-1)
  if image.ndim < 3:
    image = image[...,None]
  if touchingRoiOnly:
    out = np.zeros_like(labels, dtype=bool)
    for seed in seeds:
      out |= flood(labels, tuple(seed),)
  else:
    keepColors = labels[seeds[:,0], seeds[:,1]]
    out = np.isin(labels, keepColors)
  # Zero out negative regions from previous runs
  out[historyMask == 1] = False
  nChans = image.shape[2]
  if useMeanColor:
    summaryImg = np.zeros_like(image)
    # Offset by 1 to avoid missing 0-labels
    for lbl in regionprops(labels+1):
      coords = lbl.coords
      intensity = image[coords[:,0], coords[:,1],...].mean(0)
      summaryImg[coords[:,0], coords[:,1], :] = intensity
  else:
    if np.issubdtype(labels.dtype, np.bool_):
      labels = labels.astype('uint8')
    boundaries = labelBoundaries_cv(labels, lineThickness)
    summaryImg = image.copy()
    summaryImg[boundaries,...] = [255 for _ in range(nChans)]
  color = (255,) + tuple(0 for _ in range(1, nChans))
  cv.drawContours(summaryImg, [fgVerts], -1, color, lineThickness)
  return ProcessIO(image=out, summaryInfo={'image': summaryImg})

def region_growing(image: NChanImg, fgVerts: XYVertices, seedThresh=10):
  if image.size == 0:
    return ProcessIO(image=np.zeros(image.shape[:2], bool))
  if np.all(fgVerts == fgVerts[0, :]):
    # Remove unnecessary redundant seedpoints
    fgVerts = fgVerts[[0], :]
  # outMask = np.zeros(image.shape[0:2], bool)
  # For small enough shapes, get all boundary pixels instead of just shape vertices
  if fgVerts.connected:
    fgVerts = cornersToFullBoundary(fgVerts, 50e3)

  # Don't let region grow outside area of effect
  # img_aoe, coords = getCroppedImg(image, fgVerts, areaOfEffect, coordsAsSlices=True)
  # Offset vertices before filling
  # seeds = fgVerts - [coords[1].start, coords[0].start]
  outMask = growSeedpoint(image, fgVerts, seedThresh)

  return ProcessIO(image=outMask)

class TopLevelImageProcessors:
  @staticmethod
  def b_regionGrowProcessor():
    return ImageProcess.fromFunction(region_growing)

  @staticmethod
  def c_kMeansProcessor():
    proc = ImageProcess.fromFunction(k_means)
    # Add as process so it can be disabled
    proc.addFunction(binarize_labels)
    return proc

  @staticmethod
  def a_grabCutProcessor():
    proc = ImageProcess.fromFunction(cv_grabcut, name='Primitive Grab Cut')
    proc.addFunction(binarize_labels)
    return proc

  @staticmethod
  def ac_quickshiftProcessor():
    proc = ImageProcess.fromFunction(quickshift_seg)
    proc.addFunction(binarize_labels, ignoreKeys=['touchingRoiOnly'])
    return proc

  # @staticmethod
  # def ad_morphAcweProcessor():
  #   proc = ImageProcess.fromFunction(morph_acwe, name='Morph. ACWE Contours')
  #   proc.addFunction(binarize_labels)
  #   return proc

  # @staticmethod
  # def ab_slicProcessor():
  #   def slic(image, n_segments=100, compactness=10.0, sigma=0, min_size_factor=0.5, max_size_factor=3):
  #     return ProcessIO(labels=seg.slic(**locals(), start_label=1))
  #   slic.__doc__ = seg.slic.__doc__
  #   proc = ImageProcess.fromFunction(slic, name='SLIC')
  #   proc.addFunction(binarize_labels)
  #   return proc

  @staticmethod
  def w_basicShapesProcessor():
    def basic_shapes(image: np.ndarray, fgVerts: XYVertices):
      return ProcessIO(image=ComplexXYVertices([fgVerts]).toMask(image.shape[:2], asBool=True))
    proc = ImageProcess.fromFunction(basic_shapes, name='Basic Shapes')
    proc.disabledStages = [['Basic Region Operations', 'Open -> Close']]
    return proc


# -----
# GLOBAL PROCESSING
# -----
def get_component_images(image: np.ndarray, components: pd.DataFrame):
  """
  From a main image and dataframe of components, adds an 'img' column to `components` which holds the
  subregion within the image each component occupies.
  """
  imgs = [getCroppedImg(image, verts.stack(), 0) for verts in components[RTF.VERTICES]]
  return ProcessIO(subimages=imgs)

def dispatchedTemplateMatcher(func):
  inputSpec = ProcessIO.fromFunction(func, ignoreKeys=['template'])
  inputSpec['components'] = inputSpec.FROM_PREV_IO
  def dispatcher(image: np.ndarray, components: pd.DataFrame, **kwargs):
    out = ProcessIO()
    allComps = []
    for ii, comp in components.iterrows():
      verts = comp[RTF.VERTICES].stack()
      template = getCroppedImg(image, verts, 0, returnSlices=False)
      result = func(image=image, template=template, **kwargs)
      if isinstance(result, ProcessIO):
        pts = result['matchPts']
        out.update(**result)
        out.pop('components', None)
      else:
        pts = result
      allComps.append(pts_to_components(pts, comp))
    outComps = pd.concat(allComps, ignore_index=True)
    out['components'] = outComps
    out['deleteOrig'] = True
    return out
  dispatcher.__doc__ = func.__doc__
  proc = GlobalPredictionProcess(fns.nameFormatter(func.__name__))
  proc.addFunction(dispatcher)
  proc.stages[0].input = inputSpec
  return proc

def dispatchedFocusedProcessor(func):
  inputSpec = ProcessIO.fromFunction(func)
  inputSpec['components'] = inputSpec.FROM_PREV_IO
  def dispatcher(image: np.ndarray, components: pd.DataFrame, **kwargs):
    out = ProcessIO()
    allComps = []
    for ii, comp in components.iterrows():
      verts = comp[RTF.VERTICES].stack()
      focusedImage = getCroppedImg(image, verts, 0, returnSlices=False)
      result = func(image=focusedImage, **kwargs)
      if isinstance(result, ProcessIO):
        mask = result['image']
        out.update(**result)
        out.pop('components', None)
      else:
        mask = result
      newComp = comp.copy()
      if mask is not None:
        newComp[RTF.VERTICES] = ComplexXYVertices.fromBwMask(mask)
      allComps.append(fns.serAsFrame(newComp))
    outComps = pd.concat(allComps)
    out['components'] = outComps
    out['addType'] = PRJ_ENUMS.COMP_ADD_AS_MERGE
    return out
  dispatcher.__doc__ = func.__doc__
  proc = GlobalPredictionProcess(fns.nameFormatter(func.__name__))
  proc.addFunction(dispatcher)
  proc.stages[0].input = inputSpec
  return proc


@fns.dynamicDocstring(metricTypes=[d for d in dir(cv) if d.startswith('TM')])
def cv_template_match(template: np.ndarray, image: np.ndarray, viewbox: np.ndarray,
                      threshold=0.8, metric='TM_CCOEFF_NORMED', area='viewbox'):
  """
  Performs template matching using default opencv functions
  :param template: Template image
  :param image: Main image
  :param threshold:
    helpText: Cutoff point to consider a matched template
    limits: [0, 1]
    step: 0.1
  :param metric:
    helpText: Template maching metric
    pType: list
    limits: {metricTypes}
  :param area:
    helpText: Where to apply the new components
    pType: list
    limits:
      - image
      - viewbox
  """
  if area == 'viewbox':
    image, coords = getCroppedImg(image, viewbox, 0)
  else:
    coords = np.array([[0,0]])
  if image.ndim < 3:
    grayImg = image
  else:
    grayImg = cv.cvtColor(image, cv.COLOR_RGB2GRAY)
  if template.ndim > 2:
    template = cv.cvtColor(template, cv.COLOR_RGB2GRAY)

  metric = getattr(cv, metric)
  res = cv.matchTemplate(grayImg, template, metric)
  maxFilter = maximum_filter(res, template.shape[:2])
  # Non-max suppression to remove close-together peaks
  res[maxFilter > res] = 0
  loc = np.nonzero(res >= threshold)
  scores = res[loc]
  matchPts = np.c_[loc[::-1]] + coords[[0]]
  return ProcessIO(matchPts=matchPts, scores=scores, matchImg=maxFilter)


def pts_to_components(matchPts: np.ndarray, component: pd.Series):
  numOutComps = len(matchPts)
  # Explicit copy otherwise all rows point to the same component
  outComps = pd.concat([fns.serAsFrame(component)]*numOutComps, ignore_index=True).copy()
  origOffset = component[RTF.VERTICES].stack().min(0)
  allNewverts = []
  for ii, pt in zip(outComps.index, matchPts):
    newVerts = []
    for verts in outComps.at[ii, RTF.VERTICES]:
      newVerts.append(verts-origOffset+pt)
    allNewverts.append(ComplexXYVertices(newVerts))
  outComps[RTF.VERTICES] = allNewverts
  return outComps


@lru_cache()
def _modelFromFile(model: str):
  fname = MODEL_DIR/model
  return tf.load(fname)

def nn_model_prediction(image: NChanImg, model=''):
  """

  :param image:
  :param fgVerts:
  :param model:
    pType: list
    limits:
      - ''
      - Tool 1
      - Tool 2
      - Tool 3
  """
  if not model:
    return None
  nnModel = _modelFromFile(model)

  return np.random.random((image.shape[:2])) > 0

TOP_GLOBAL_PROCESSOR_FUNCS = [
  lambda: dispatchedTemplateMatcher(cv_template_match),
  # lambda: dispatchedFocusedProcessor(nn_model_prediction)
]


import warnings
from collections import defaultdict
from pathlib import Path
from typing import Union, Dict, List

import numpy as np
import pyqtgraph as pg
import qdarkstyle
from pandas import DataFrame as df
from pyqtgraph.Qt import QtCore, QtWidgets, QtGui
from utilitys import ParamEditor, ParamEditorPlugin, RunOpts, PrjParam, fns, \
  ParamEditorDockGrouping

from s3a.generalutils import hierarchicalUpdate
from s3a.constants import LAYOUTS_DIR, REQD_TBL_FIELDS, ICON_DIR
from s3a.constants import PRJ_ENUMS
from s3a.models.s3abase import S3ABase
from s3a.parameditors import PRJ_SINGLETON
from s3a.plugins.file import FilePlugin
from s3a.plugins.misc import RandomToolsPlugin, MainImagePlugin, CompTablePlugin
from s3a.structures import XYVertices, FilePath, NChanImg

__all__ = ['S3A']

_MENU_PLUGINS = [RandomToolsPlugin]

class S3A(S3ABase):
  sigLayoutSaved = QtCore.Signal()
  S3A_INST = None

  __groupingName__ = 'Application'

  @classmethod
  def __initEditorParams__(cls):
    super().__initEditorParams__()
    cls.toolsEditor = ParamEditor.buildClsToolsEditor(cls, 'Main Window')

  def __init__(self, parent=None, guiMode=True, loadLastState=True, **startupSettings):
    # Wait to import quick loader profiles until after self initialization so
    # customized loading functions also get called
    super().__init__(parent, **startupSettings)
    self.setWindowIcon(QtGui.QIcon(str(ICON_DIR/'s3alogo.svg')))
    if guiMode:
      warnings.simplefilter('error', UserWarning)
      fns.makeExceptionsShowDialogs(self)
    self.APP_TITLE = 'FICS Semi-Supervised Semantic Annotator'
    self.CUR_COMP_LBL = 'Current Component ID:'
    self.setWindowTitle(self.APP_TITLE)
    self.setWindowIconText(self.APP_TITLE)

    self.curCompIdLbl = QtWidgets.QLabel(self.CUR_COMP_LBL)

    # -----
    # LAOYUT MANAGER
    # -----
    # Dummy editor for layout options since it doesn't really have editable settings
    # Maybe later this can be elevated to have more options
    self.layoutEditor = ParamEditor(self, None, LAYOUTS_DIR, 'dockstate', 'Layout')
    def loadLayout(layoutName: Union[str, Path]):
      layoutName = Path(layoutName)
      if not layoutName.is_absolute():
        layoutName = LAYOUTS_DIR/f'{layoutName}.dockstate'
      self.restoreState(fns.attemptFileLoad(layoutName))

    def saveRecentLayout(_folderName: Path):
      outFile = _folderName/'savedLayout.dockstate'
      self.saveLayout(outFile)
      return str(outFile)

    self.layoutEditor.loadParamValues = loadLayout
    self.layoutEditor.saveParamValues = saveRecentLayout
    self.appStateEditor.addImportExportOpts('layout', loadLayout, saveRecentLayout)

    # When docks are loaded via layout editor, they don't resize properly. Fix that here
    for dock in PRJ_SINGLETON.docks:
      editors = dock.editors if isinstance(dock, ParamEditorDockGrouping) else [dock]
      for editor in editors:
        editor.tree.resizeColumnToContents(0)

    self._buildGui()
    self._buildMenu()
    self._hookupSignals()

    # Load in startup settings
    stateDict = None if loadLastState else {}
    hierarchicalUpdate(self.appStateEditor.startupSettings, startupSettings)
    self.appStateEditor.loadParamValues(stateDict=stateDict)

  def _hookupSignals(self):
    PRJ_SINGLETON.colorScheme.registerFunc(self.updateTheme, runOpts=RunOpts.ON_CHANGED, nest=False)
    # EDIT
    self.saveAllEditorDefaults()

  def _buildGui(self):
    self.setDockOptions(self.ForceTabbedDocks)
    self.setTabPosition(QtCore.Qt.AllDockWidgetAreas, QtWidgets.QTabWidget.North)
    centralWidget = QtWidgets.QWidget()
    self.setCentralWidget(centralWidget)
    layout = QtWidgets.QVBoxLayout(centralWidget)

    self.toolbarWidgets: Dict[PrjParam, List[QtWidgets.QAction]] = defaultdict(list)
    layout.addWidget(self.mainImg)

    self.tblFieldToolbar.setObjectName('Table Field Plugins')
    self.addToolBar(self.tblFieldToolbar)
    self.generalToolbar.setObjectName('General')
    self.addToolBar(self.generalToolbar)

    _plugins = [PRJ_SINGLETON.clsToPluginMapping[c] for c in [MainImagePlugin, CompTablePlugin]]
    parents = [self.mainImg, self.compTbl]
    for plugin, parent in zip(_plugins, reversed(parents)):
      plugin.toolsEditor.actionsMenuFromProcs(plugin.name, nest=True, parent=parent, outerMenu=parent.menu)


    tableDock = QtWidgets.QDockWidget('Component Table Window', self)
    tableDock.setFeatures(tableDock.DockWidgetMovable|tableDock.DockWidgetFloatable)

    tableDock.setObjectName('Component Table Dock')
    tableContents = QtWidgets.QWidget(tableDock)
    tableLayout = QtWidgets.QVBoxLayout(tableContents)
    tableLayout.addWidget(self.compTbl)
    tableDock.setWidget(tableContents)

    self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, tableDock)

    # STATUS BAR
    self.setStatusBar(self.statBar)

    self.mouseCoordsLbl = QtWidgets.QLabel()
    self.pxColorLbl = QtWidgets.QLabel()

    self.imageLbl = QtWidgets.QLabel(f"Image: None")

    self.statBar.show()
    self.statBar.addWidget(self.imageLbl)

    self.statBar.addWidget(self.mouseCoordsLbl)
    self.mainImg.mouseCoordsLbl = self.mouseCoordsLbl

    self.statBar.addWidget(self.pxColorLbl)
    self.mainImg.pxColorLbl = self.pxColorLbl

  def saveLayout(self, layoutName: Union[str, Path]=None, allowOverwriteDefault=False):
    dockStates = self.saveState().data()
    if Path(layoutName).is_absolute():
      savePathPlusStem = layoutName
    else:
      savePathPlusStem = LAYOUTS_DIR/layoutName
    saveFile = savePathPlusStem.with_suffix(f'.dockstate')
    fns.saveToFile(dockStates, saveFile,
                   allowOverwriteDefault=allowOverwriteDefault)
    self.sigLayoutSaved.emit()

  def changeFocusedComp(self, newComps: df=None, forceKeepLastChange=False):
    ret = super().changeFocusedComp(newComps, forceKeepLastChange)
    self.curCompIdLbl.setText(f'Component ID: {self.mainImg.compSer[REQD_TBL_FIELDS.INST_ID]}')
    return ret

  def resetTblFields_gui(self):
    outFname = fns.popupFilePicker(None, 'Select Table Config File', 'All Files (*.*);; Config Files (*.yml)')
    if outFname is not None:
      PRJ_SINGLETON.tableData.loadCfg(outFname)

  def _buildMenu(self):
    # TODO: Find a better way of fixing up menu order
    menus = self.menuBar_.actions()
    menuFile = [a for a in menus if a.text() == FilePlugin.name][0]
    self.menuBar_.insertAction(menus[0], menuFile)

  def _handleNewPlugin(self, plugin: ParamEditorPlugin):
    super()._handleNewPlugin(plugin)
    dock = plugin.dock
    if dock is None:
      return
    PRJ_SINGLETON.quickLoader.addDock(dock)
    self.addTabbedDock(QtCore.Qt.RightDockWidgetArea, dock)

    if plugin.menu is None:
      # No need to add menu and graphics options
      return

    parentTb = plugin.parentMenu
    if parentTb is not None:
      self.createMenuOptForPlugin(plugin, parentToolbarOrMenu=parentTb)

  def setMainImg(self, fileName: FilePath = None, imgData: NChanImg = None,
                 clearExistingComps=True):
    ret = super().setMainImg(fileName, imgData, clearExistingComps)
    img = self.srcImgFname
    if img is not None:
      img = img.name
    self.imageLbl.setText(f'Image: {img}')

    return ret

  def setMainImg_gui(self):
    fileFilter = "Image Files (*.png *.tif *.jpg *.jpeg *.bmp *.jfif);;All files(*.*)"
    fname = fns.popupFilePicker(None, 'Select Main Image', fileFilter)
    if fname is not None:
      with pg.BusyCursor():
        self.setMainImg(fname)

  def exportAnnotations_gui(self):
    """Saves the component table to a file"""
    fileFilters = self.compIo.handledIoTypes_fileFilter(**{'*': 'All Files'})
    outFname = fns.popupFilePicker(None, 'Select Save File', fileFilters, existing=False)
    if outFname is not None:
      super().exportCurAnnotation(outFname)

  def openAnnotation_gui(self):
    # TODO: See note about exporting comps. Delegate the filepicker activity to importer
    fileFilter = self.compIo.handledIoTypes_fileFilter()
    fname = fns.popupFilePicker(None, 'Select Load File', fileFilter)
    if fname is None:
      return
    self.openAnnotations(fname)

  def saveLayout_gui(self):
    outName = fns.dialogGetSaveFileName(self, 'Layout Name')
    if outName is None or outName == '':
      return
    self.saveLayout(outName)

  # ---------------
  # BUTTON CALLBACKS
  # ---------------
  def closeEvent(self, ev: QtGui.QCloseEvent):
    # Confirm all components have been saved
    shouldExit = False
    if self.hasUnsavedChanges:
      ev.ignore()
      if (QtWidgets.QMessageBox().question(self, 'Confirm Exit',
                                         'Component table has unsaved changes.\nAre you sure you want to exit?',
                                         QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Cancel)
          == QtWidgets.QMessageBox.Ok):
        shouldExit = True
    else:
      shouldExit = True
    if shouldExit:
      # Clean up all editor windows, which could potentially be left open
      ev.accept()
      PRJ_SINGLETON.close()
      fns.restoreExceptionBehavior()
      self.appStateEditor.saveParamValues()

  def forceClose(self):
    """
    Allows the app to close even if it has unsaved changes. Useful for closing
    within a script
    """
    self.hasUnsavedChanges = False
    self.close()

  def createMenuOptForPlugin(self, plugin: ParamEditorPlugin, parentToolbarOrMenu=None):
    if isinstance(parentToolbarOrMenu, QtWidgets.QToolBar):
      btn = QtWidgets.QToolButton()
      btn.setText(plugin.name)
      parentToolbarOrMenu.addWidget(btn)
      parentToolbarOrMenu = btn
      btn.addMenu = btn.setMenu
      btn.addAction = lambda act: act.triggered.connect(btn.click)
      btn.setPopupMode(btn.InstantPopup)
    parentToolbarOrMenu.addMenu(plugin.menu)
    oldShow = plugin.dock.showEvent
    def show_fixDockWidth(ev):
      oldShow(ev)
      dock = plugin.dock
      dock.raise_()
      dock.activateWindow()
      if dock.width() < dock.biggestMinWidth + 100:
        self.resizeDocks([dock], [dock.biggestMinWidth + 100], QtCore.Qt.Horizontal)
    plugin.dock.showEvent = show_fixDockWidth

  def _populateLoadLayoutOptions(self):
    self.layoutEditor.addDirItemsToMenu(self.menuLayout)

  def updateTheme(self, useDarkTheme=False):
    style = ''
    if useDarkTheme:
      style = qdarkstyle.load_stylesheet()
    self.setStyleSheet(style)

  def add_focusComps(self, newComps: df, addType=PRJ_ENUMS.COMP_ADD_AS_NEW):
    ret = super().add_focusComps(newComps, addType=addType)
    selection = self.compDisplay.selectRowsById(newComps[REQD_TBL_FIELDS.INST_ID])
    if self.isVisible() and self.compTbl.showOnCreate:
      # For some reason sometimes the actual table selection doesn't propagate in time, so
      # directly forward the selection here
      self.compTbl.setSelectedCellsAs_gui(selection)
    return ret

if __name__ == '__main__':
  import sys
  app = QtWidgets.QApplication([])
  win = S3A()
  win.showMaximized()
  sys.exit(app.exec_())


import sys
from pathlib import Path
from typing import Optional, Union, Type
from warnings import warn

import numpy as np
import pandas as pd
from pandas import DataFrame as df
from pyqtgraph.Qt import QtCore, QtWidgets

from s3a.generalutils import hierarchicalUpdate
from utilitys import EditorPropsMixin, RunOpts, ParamEditorPlugin, fns, ParamEditor

from s3a import ComponentIO
from s3a.constants import PRJ_CONSTS, REQD_TBL_FIELDS, LAYOUTS_DIR
from s3a.constants import PRJ_ENUMS
from s3a.controls.tableviewproxy import CompDisplayFilter, CompSortFilter
from s3a.models.tablemodel import ComponentMgr
from s3a.parameditors import PRJ_SINGLETON
from s3a.parameditors.appstate import AppStateEditor
from s3a.plugins.file import FilePlugin
from s3a.structures import FilePath, NChanImg
from s3a.views.imageareas import MainImage
from s3a.views.tableview import CompTableView

__all__ = ['S3ABase']

class S3ABase(EditorPropsMixin, QtWidgets.QMainWindow):
  """
  Top-level widget for producing component bounding regions from an input image.
  """

  sigRegionAccepted = QtCore.Signal()

  __groupingName__ = 'S3A Window'

  def __init__(self, parent=None, **startupSettings):
    super().__init__(parent)
    self.tblFieldToolbar = QtWidgets.QToolBar('Table Field Plugins')
    self.generalToolbar = QtWidgets.QToolBar('General')

    self.mainImg = MainImage(toolbar=self.generalToolbar)
    PRJ_CONSTS.TOOL_ACCEPT_FOC_REGION.opts['ownerObj'] = self.mainImg
    self.mainImg.toolsEditor.registerFunc(self.acceptFocusedRegion,
                                          btnOpts=PRJ_CONSTS.TOOL_ACCEPT_FOC_REGION)
    _, param = PRJ_SINGLETON.generalProps.registerFunc(PRJ_SINGLETON.actionStack.resizeStack,
                                                       name=self.__groupingName__,
                                                       runOpts=RunOpts.ON_CHANGED,
                                                       maxLength=PRJ_CONSTS.PROP_UNDO_BUF_SZ.value,
                                                       returnParam=True)
    param.child('maxLength').setOpts(title=PRJ_CONSTS.PROP_UNDO_BUF_SZ.name)
    self.statBar = QtWidgets.QStatusBar(self)
    self.menuBar_ = self.menuBar()

    PRJ_SINGLETON.shortcuts.registerShortcut(PRJ_CONSTS.TOOL_CLEAR_ROI,
                                             self.mainImg.clearCurRoi,
                                             overrideOwnerObj=self.mainImg
                                             )

    self.compMgr = ComponentMgr()
    # Register exporter to allow user parameters
    ioCls = ComponentIO
    with PRJ_SINGLETON.generalProps.setBaseRegisterPath(PRJ_CONSTS.CLS_COMP_EXPORTER.name):
      ioCls.exportOnlyVis, ioCls.includeFullSourceImgName = \
        PRJ_SINGLETON.generalProps.registerProps(
          [PRJ_CONSTS.EXP_ONLY_VISIBLE, PRJ_CONSTS.INCLUDE_FNAME_PATH])
    self.compIo: ComponentIO = ioCls()
    ComponentIO.tableData = PRJ_SINGLETON.tableData

    self.compTbl = CompTableView()
    self.compDisplay = CompDisplayFilter(self.compMgr, self.mainImg, self.compTbl)

    self.compTbl.setSortingEnabled(True)
    self.compTbl.setAlternatingRowColors(True)
    # Allow filtering/sorting
    self.sortFilterProxy = CompSortFilter(self.compMgr)
    self.compTbl.setModel(self.sortFilterProxy)

    self.hasUnsavedChanges = False
    self.srcImgFname: Optional[Path] = None

    self.appStateEditor = AppStateEditor(self, name='App State Editor')

    # -----
    # INTERFACE WITH QUICK LOADER / PLUGINS
    # -----
    PRJ_SINGLETON.tableData.sigCfgUpdated.connect(lambda: self.resetTblFields())
    self.filePlg: FilePlugin = self.addPlugin(FilePlugin)

    for plugin in PRJ_SINGLETON.clsToPluginMapping.values(): # type: ParamEditorPlugin
      # Plugins created before window was initialized may need their plugins forcefully
      # attached here
      if plugin.win is not self:
        self._handleNewPlugin(plugin)
    PRJ_SINGLETON.sigPluginAdded.connect(self._handleNewPlugin)

    # Connect signals
    # -----
    # COMPONENT MANAGER
    # -----
    def handleUpdate(*_args):
      self.hasUnsavedChanges = True
    self.compMgr.sigCompsChanged.connect(handleUpdate)

    # -----
    # MAIN IMAGE
    # -----
    def handleCompsChanged(changedDict: dict):
      ser = self.mainImg.compSer
      focusedId = ser[REQD_TBL_FIELDS.INST_ID]
      if focusedId in changedDict['deleted']:
        self.changeFocusedComp()
      elif focusedId in changedDict['changed']:
        ser.update(self.compMgr.compDf.loc[focusedId])
    self.compMgr.sigCompsChanged.connect(handleCompsChanged)

    # -----
    # COMPONENT TABLE
    # -----
    self.compDisplay.sigCompsSelected.connect(lambda newComps: self.changeFocusedComp(newComps))

    # -----
    # MISC
    # -----
    self.saveAllEditorDefaults()
    PRJ_SINGLETON.tableData.sigCfgUpdated.connect(lambda: self.resetTblFields())

  def resetTblFields(self):
    """
    When table fields change, the displayed columns must change and the view
    must be made aware. Ensure this occurs here
    """
    # Even if the field names are the same, e.g. classes may added or default values could
    # be changed. So, reset the cell editor delegates no matter what
    self.compTbl.setColDelegates()
    self.compTbl.popup.reflectDelegateChange()
    # Make sure this is necessary, first
    for mgr in self.compMgr, self.compTbl.popup.tbl.mgr:
      if mgr.colTitles == list([f.name for f in PRJ_SINGLETON.tableData.allFields]):
        # Fields haven't changed since last reset. Types could be different, but nothing
        # will break. So, the table doesn't have to be completely reset
        return

      mgr.beginResetModel()
      mgr.rmComps()
      mgr.resetFields()
      mgr.endResetModel()

  def _handleNewPlugin(self, plugin: ParamEditorPlugin):
    plugin.attachWinRef(self)
    if plugin.dock:
      plugin.dock.setParent(self)

  @staticmethod
  def saveAllEditorDefaults():
    for editor in PRJ_SINGLETON.registerableEditors:
      editor.saveCurStateAsDefault()

  @PRJ_SINGLETON.actionStack.undoable('Accept Focused Region')
  def acceptFocusedRegion(self):
    """Applies the focused image vertices to the corresponding component in the table"""
    # If the component was deleted
    mgr = self.compMgr
    focusedId = self.mainImg.compSer[REQD_TBL_FIELDS.INST_ID]
    exists = focusedId in mgr.compDf.index
    if not exists and focusedId != REQD_TBL_FIELDS.INST_ID.value:
      # Could be a brand new component, allow in that case
      warn('Cannot accept region as this component was deleted.', UserWarning)
      return

    self.sigRegionAccepted.emit()

    ser = self.mainImg.compSer
    if ser[REQD_TBL_FIELDS.VERTICES].isEmpty():
      # Component should be erased. Since new components will not match existing
      # IDs the same function will work regardless of whether this was new or existing
      self.compMgr.rmComps([ser[REQD_TBL_FIELDS.INST_ID]])
      return

    if exists:
      undo = self._acceptFocused_existing(ser)
    else:
      undo = self._acceptFocused_new(ser)
    self.changeFocusedComp()
    yield
    undo()


  def _acceptFocused_new(self, compSer: pd.Series):
    # New, make a brand new table entry
    newIds = self.compMgr.addComps(fns.serAsFrame(compSer))['added']
    def undo():
      self.compMgr.rmComps(newIds)
    return undo

  def _acceptFocused_existing(self, compSer: pd.Series):
    oldComp = self.compMgr.compDf.loc[[compSer[REQD_TBL_FIELDS.INST_ID]]].copy()
    modifiedDf = fns.serAsFrame(compSer)
    self.compMgr.addComps(modifiedDf, addtype=PRJ_ENUMS.COMP_ADD_AS_MERGE)
    def undo():
      self.add_focusComps(oldComp, addType=PRJ_ENUMS.COMP_ADD_AS_MERGE)
    return undo


  def clearBoundaries(self):
    """Removes all components from the component table"""
    self.compMgr.rmComps()

  def addPlugin(self, pluginCls: Type[ParamEditorPlugin], *args, **kwargs):
    """See PRJ_SINGLETON.addPlugin"""
    plugin = PRJ_SINGLETON.addPlugin(pluginCls, *args, **kwargs)
    return plugin

  @PRJ_SINGLETON.actionStack.undoable('Change Main Image')
  def setMainImg(self, fileName: FilePath=None, imgData: NChanImg=None,
                 clearExistingComps=True):
    """
    * If fileName is None, the main and focused images are blacked out.
    * If only fileName is provided, it is assumed to be an image. The image data
    will be populated by reading in that file.
    * If both fileName and imgData are provided, then imgData is used to populate the
    image, and fileName is assumed to be the file associated with that data.

    :param fileName: Filename either to load or that corresponds to imgData
    :param imgData: N-Channel numpy image
    :param clearExistingComps: If True, erases all existing components on image load.
      Else, they are retained.
    """
    oldFile = self.srcImgFname
    oldData = self.mainImg.image
    if fileName is not None:
      fileName = Path(fileName).resolve()
    if fileName == self.srcImgFname:
      return

    self.saveCurAnnotation()

    if imgData is not None:
      self.mainImg.setImage(imgData)
    else:
      self.mainImg.setImage(fileName)
    self.srcImgFname = fileName

    self.clearBoundaries()
    self.mainImg.plotItem.vb.autoRange()
    if fileName is not None:
      self.filePlg.addImage(fileName)
    self.loadNewAnnotations()
    yield
    self.setMainImg(oldFile, oldData, clearExistingComps)

  def saveCurAnnotation(self, allowOverwrite=False):
    srcImg = self.srcImgFname
    if srcImg is None:
      return
    srcImg_proj = self.filePlg.imagesDir/srcImg.name
    if srcImg_proj.exists() and not allowOverwrite and srcImg_proj != srcImg:
      # warn(f'Image {srcImg.name} already present in project images. Using'
      #      f' existing image instead of current data.', UserWarning)
      pass
    else:
      self.filePlg.addImage(srcImg, data=self.mainImg.image, copyToProj=True,
                            allowOverwrite=True)
    # srcImg_proj is guaranteed to exist at this point
    self.filePlg.addAnnotation(data=self.exportableDf, image=srcImg_proj, overwriteOld=True)
    self.srcImgFname = srcImg_proj
    self.hasUnsavedChanges = False

  def loadNewAnnotations(self, imgFname: FilePath=None):
    if imgFname is None:
      imgFname = self.srcImgFname
    if imgFname is None:
      return
    imgAnns = self.filePlg.imgToAnnMapping.get(imgFname, None)
    if imgAnns is not None:
      self.compMgr.addComps(self.compIo.buildByFileType(imgAnns, imgDir=self.filePlg.imagesDir,
                                                        imShape=self.mainImg.image.shape))
      # 'hasUnsavedChanges' will be true after this, even though the changes are saved.
      self.hasUnsavedChanges = False

  @fns.dynamicDocstring(filters=ComponentIO.handledIoTypes_fileFilter())
  def exportCurAnnotation(self, outFname: Union[str, Path], readOnly=True, verifyIntegrity=True):
    """
    Exports current image annotations to a file. This may be more convenient than exporting
    an entire project if just the current current annotations are needed

    :param outFname:
      title: Output File
      helpText: Where to export. The file extension determines the save type
      pType: filepicker
      existing: False
      fileFilter: {filters}
    :param readOnly: Whether the export should be read-only
    :param verifyIntegrity: Whether the export should be re-imported to confirm
      its contents were successfully exported. This is only applicable for file types
      in `ComponentIO.roundTripIoTypes`
    """
    self.compIo.exportByFileType(self.exportableDf, outFname, imShape=self.mainImg.image.shape,
                                 readOnly=readOnly, verifyIntegrity=verifyIntegrity)
  @property
  def exportableDf(self):
    """
    Dataframe from manager with populated information for main image name and
    potentially filtered to only visible components (if requested by the user)
    """
    displayIds = self.compDisplay.displayedIds
    srcImgFname = self.srcImgFname
    if self.compIo.exportOnlyVis and displayIds is not None:
      exportIds = displayIds
    else:
      exportIds = self.compMgr.compDf.index
    exportDf: df = self.compMgr.compDf.loc[exportIds].copy()
    if not self.compIo.includeFullSourceImgName and srcImgFname is not None:
      # Only use the file name, not the whole path
      srcImgFname = srcImgFname.name
    elif srcImgFname is not None:
      srcImgFname = str(srcImgFname)
    # Assign correct export name for only new components
    overwriteIdxs = exportDf[REQD_TBL_FIELDS.SRC_IMG_FILENAME] == PRJ_CONSTS.ANN_CUR_FILE_INDICATOR.value
    # TODO: Maybe the current filename will match the current file indicator. What happens then?
    exportDf.loc[overwriteIdxs, REQD_TBL_FIELDS.SRC_IMG_FILENAME] = srcImgFname
    return exportDf

  def openAnnotations(self, inFname: str, loadType=PRJ_ENUMS.COMP_ADD_AS_NEW):
    pathFname = Path(inFname)
    if self.mainImg.image is None:
      raise IOError('Cannot load components when no main image is set.')
    fType = pathFname.suffix[1:]
    if not any(fType in typ for typ in self.compIo.handledIoTypes):
      raise IOError(f'Extension {fType} is not recognized. Must be one of:\n'
                       + self.compIo.handledIoTypes_fileFilter())
    newComps = self.compIo.buildByFileType(inFname, self.mainImg.image.shape)
    self.compMgr.addComps(newComps, loadType)

  @PRJ_SINGLETON.actionStack.undoable('Create New Comp', asGroup=True)
  def add_focusComps(self, newComps: df, addType=PRJ_ENUMS.COMP_ADD_AS_NEW):
    changeDict = self.compMgr.addComps(newComps, addType)
    # Focus is performed by comp table
    # Arbitrarily choose the last possible component
    changeList = np.concatenate([changeDict['added'], changeDict['changed']])
    if len(changeList) == 0:
      return
    self.changeFocusedComp(self.compMgr.compDf.loc[[changeList[-1]]])

  @PRJ_SINGLETON.actionStack.undoable('Change Focused Component')
  def changeFocusedComp(self, newComps: df=None, forceKeepLastChange=False):
    oldSer = self.mainImg.compSer.copy()
    oldImg = self.mainImg.image
    if newComps is None or len(newComps) == 0:
      self.mainImg.updateFocusedComp()
      self.compDisplay.regionPlot.focusById([])
      self.compDisplay.selectRowsById([])
    else:
      # TODO: More robust scenario if multiple comps are in the dataframe
      #   For now, just use the last in the selection. This is so that if multiple
      #   components are selected in a row, the most recently selected is always
      #   the current displayed.
      newComp: pd.Series = newComps.iloc[-1,:]
      newCompId = newComp[REQD_TBL_FIELDS.INST_ID]
      self.compDisplay.regionPlot.focusById([newCompId])
      self.mainImg.updateFocusedComp(newComp)
    # Nothing happened since the last component change, so just replace it instead of
    # adding a distinct action to the buffer queue
    stack = PRJ_SINGLETON.actionStack
    if not forceKeepLastChange and stack.undoDescr == 'Change Focused Component':
      stack.actions.pop()
    yield
    if oldImg is not None and len(oldSer.loc[REQD_TBL_FIELDS.VERTICES]) > 0:
      self.changeFocusedComp(fns.serAsFrame(oldSer), forceKeepLastChange=True)
    else:
      self.changeFocusedComp()

  # Stolen and adapted for python from https://stackoverflow.com/a/42910109/9463643
  # noinspection PyTypeChecker
  def addTabbedDock(self, area: QtCore.Qt.DockWidgetArea, dockwidget: QtWidgets.QDockWidget):
    curAreaWidgets = [d for d in self.findChildren(QtWidgets.QDockWidget)
                      if self.dockWidgetArea(d) == area]
    try:
      self.tabifyDockWidget(curAreaWidgets[-1], dockwidget)
    except IndexError:
      # First dock in area
      self.addDockWidget(area, dockwidget)


from functools import lru_cache
from pathlib import Path
from typing import Any, Optional, Callable, Tuple, Union, Sequence, List, Collection
from contextlib import contextmanager

import cv2 as cv
import numpy as np
from pandas import DataFrame as df
from utilitys import PrjParam
from utilitys.typeoverloads import FilePath

from .structures import TwoDArr, XYVertices, ComplexXYVertices, NChanImg, BlackWhiteImg


def stackedVertsPlusConnections(vertList: ComplexXYVertices) -> (XYVertices, np.ndarray):
  """
  Utility for concatenating all vertices within a list while recording where separations
  occurred
  """
  allVerts = [np.zeros((0,2))]
  separationIdxs = []
  idxOffset = 0
  for curVerts in vertList:
    allVerts.append(curVerts)
    vertsLen = len(curVerts)
    if vertsLen == 0: continue
    # Close the current shape
    allVerts.append(curVerts[0,:])
    separationIdxs.append(idxOffset + vertsLen)
    idxOffset += vertsLen + 1
  # Take away last separator if it exists
  if len(separationIdxs) > 0:
    separationIdxs.pop()
  allVerts = np.vstack(allVerts)
  isfinite = np.ones(len(allVerts), bool)
  isfinite[separationIdxs] = False
  return XYVertices(allVerts, dtype=float), isfinite
  #return XYVertices(dtype=float)


def splitListAtNans(concatVerts:XYVertices):
  """
  Utility for taking a single list of nan-separated region vertices
  and breaking it into several regions with no nans.
  """
  allVerts = []
  nanEntries = np.nonzero(np.isnan(concatVerts[:,0]))[0]
  curIdx = 0
  for nanEntry in nanEntries:
    curVerts = concatVerts[curIdx:nanEntry,:].astype('int')
    allVerts.append(curVerts)
    curIdx = nanEntry+1
  # Account for final grouping of verts
  allVerts.append(concatVerts[curIdx:,:].astype('int'))
  return ComplexXYVertices(allVerts, coerceListElements=True)

def getClippedBbox(arrShape: tuple, bbox: TwoDArr, margin: int):
  """
  Given a bounding box and margin, create a clipped bounding box that does not extend
  past any dimension size from arrShape

  Parameters
  ----------
  arrShape :    2-element tuple
     Refrence array dimensions

  bbox     :    2x2 array
     [minX minY; maxX maxY] bounding box coordinates

  margin   :    int
     Offset from bounding box coords. This will not fully be added to the bounding box
     if the new margin causes coordinates to fall off either end of the reference array shape.
  """
  bbox = bbox.astype(int)
  bbox[0] -= margin
  # Add extra 1 since python slicing stops before last index
  bbox[1] += margin+1
  arrShape = arrShape[:2]
  return np.clip(bbox, 0, arrShape[::-1])

def coerceDfTypes(dataframe: df, constParams: Collection[PrjParam]=None):
  """
  Pandas currently has a bug where datatypes are not preserved after update operations.
  Current workaround is to coerce all types to their original values after each operation
  """
  if constParams is None:
    constParams = dataframe.columns
  for field in constParams:
    try:
      dataframe[field] = dataframe[field].astype(type(field.value))
    except TypeError:
      # Coercion isn't possible, nothing to do here
      pass
  return dataframe

def largestList(verts: List[XYVertices]) -> XYVertices:
  maxLenList = []
  for vertList in verts:
    if len(vertList) > len(maxLenList): maxLenList = vertList
  # for vertList in newVerts:
  # vertList += cropOffset[0:2]
  return XYVertices(maxLenList)

def augmentException(ex: Exception, prependedMsg: str):
  ex.args = (prependedMsg + str(ex),)

def makeUniqueBaseClass(obj: Any):
  """
  Overwrites obj's class to a mixin base class.
  Property objects only work in Python if assigned to the *class* of an object, e.g.
  >>> class b:
  >>>   num = property(lambda self: 4)
  >>> ob = b()
  >>> ob.num # works as expected
  >>> ob.num2 = property(lambda self: 6)
  >>> ob.num2 # Property object at ... -- NOT AS EXPECTED!
  To work around this, simply use <type(ob).num2 = ...>. However, for regisetering properties,
  this means *all* objects of that type will have the properties of all other objects.
  To fix this, a mixin is added to this object and the property is added to the mixin.
  That way, the original object class is not altered, and each separate object will not
  end up sharing the same parameters.
  In summary, this feature enables the assignment
  >>> type(ob).a = property(...)
  without all other `b` objects also receiving this property.
  """
  class mixin(type(obj)): pass
  obj.__class__ = mixin
  return mixin


def lower_NoSpaces(name: str):
  return name.replace(' ', '').lower()

def safeCallFuncList(fnNames: Collection[str], funcLst: List[Callable],
                     fnArgs: List[Sequence]=None):
  errs = []
  rets = []
  if fnArgs is None:
    fnArgs = [()]*len(fnNames)
  for key, fn, args in zip(fnNames, funcLst, fnArgs):
    curRet, curErr = safeCallFunc(key, fn, *args)
    rets.append(curRet)
    if curErr: errs.append(curErr)
  return rets, errs

def safeCallFunc(fnName: str, func: Callable, *fnArgs):
  ret = err = None
  try:
    ret = func(*fnArgs)
  except Exception as ex:
    err = f'{fnName}: {ex}'
  return ret, err

def _maybeBgrToRgb(image: np.ndarray):
  """Treats 3/4-channel images as BGR/BGRA for opencv saving/reading"""
  if image.ndim > 2:
    if image.shape[0] == 1:
      image = image[...,0]
    if image.shape[2] >= 3:
      lastAx = np.arange(image.shape[2], dtype='int')
      # Swap B & R
      lastAx[[0,2]] = [2,0]
      image = image[...,lastAx]
  return image

def cvImsave_rgb(fname: FilePath, image: np.ndarray, *args, **kwargs):
  image = _maybeBgrToRgb(image)
  cv.imwrite(str(fname), image, *args, **kwargs)

def cvImread_rgb(fname: FilePath, *args, **kwargs):
  image = cv.imread(str(fname), *args, **kwargs)
  return _maybeBgrToRgb(image)


def cornersToFullBoundary(cornerVerts: Union[XYVertices, ComplexXYVertices], sizeLimit: float=np.inf,
                          fillShape: Tuple[int]=None, stackResult=True) -> Union[XYVertices, ComplexXYVertices]:
  """
  From a list of corner vertices, returns a list with one vertex for every border pixel.
  Example:
  >>> cornerVerts = XYVertices([[0,0], [100,0], [100,100],[0,100]])
  >>> cornersToFullBoundary(cornerVerts)
  # [[0,0], [1,0], ..., [100,0], [100,1], ..., [100,100], ..., ..., [0,100]]
  :param cornerVerts: Corners of the represented polygon
  :param sizeLimit: The largest number of pixels from the enclosed area allowed before the full boundary is no
  longer returned. For instance:
    >>> cornerVerts = XYVertices([[0,0], [1000,0], [1000,1000],[0,1000]])
    >>> cornersToFullBoundary(cornerVerts, 10e5)
    will *NOT* return all boundary vertices, since the enclosed area (10e6) is larger than sizeLimit.
  :param fillShape: Size of mask to create. Useful if verts may extend beyond image dimensions
    and should be truncated. If None, no truncation will occur except for negative verts.
  :param stackResult: Whether the result should be ComplexXYVertices (if stackResult is False)
    or a stacked list of exterior verts (if stackResult is True)
  :return: List with one vertex for every border pixel, unless *sizeLimit* is violated.
  """
  if isinstance(cornerVerts, XYVertices):
    cornerVerts = ComplexXYVertices([cornerVerts])
  if fillShape is not None:
    fillShape = tuple(fillShape)
  filledMask = cornerVerts.toMask(fillShape, warnIfTooSmall=False)
  cornerVerts = ComplexXYVertices.fromBwMask(filledMask, simplifyVerts=False)
  if not stackResult:
    return cornerVerts
  cornerVerts = cornerVerts.filledVerts().stack()
  numCornerVerts = len(cornerVerts)
  if numCornerVerts > sizeLimit:
    spacingPerSamp = int(numCornerVerts/sizeLimit)
    cornerVerts = cornerVerts[::spacingPerSamp]
  return cornerVerts


def getCroppedImg(image: NChanImg, verts: np.ndarray, margin=0, *otherBboxes: np.ndarray,
                  coordsAsSlices=False, returnSlices=True) -> (np.ndarray, np.ndarray):
  verts = np.vstack(verts)
  img_np = image
  compCoords = np.vstack([verts.min(0), verts.max(0)])
  if len(otherBboxes) > 0:
    for dim in range(2):
      for ii, cmpFunc in zip(range(2), [min, max]):
        otherCmpVals = [curBbox[ii, dim] for curBbox in otherBboxes]
        compCoords[ii,dim] = cmpFunc(compCoords[ii,dim], *otherCmpVals)
  compCoords = getClippedBbox(img_np.shape, compCoords, margin)
  coordSlices = (slice(compCoords[0,1], compCoords[1,1]),
                 slice(compCoords[0,0],compCoords[1,0]))
  # Verts are x-y, index into image with row-col
  indexer = coordSlices
  if image.ndim > 2:
    indexer += (slice(None),)
  croppedImg = image[indexer]
  if not returnSlices:
    return croppedImg
  if coordsAsSlices:
    return croppedImg, coordSlices
  else:
    return croppedImg, compCoords

def imgCornerVertices(img: NChanImg=None):
  """Returns [x,y] vertices for each corner of the input image"""
  if img is None:
    return XYVertices()
  fullImShape_xy = img.shape[:2][::-1]
  return XYVertices([[0,                   0],
              [0,                   fullImShape_xy[1]-1],
              [fullImShape_xy[0]-1, fullImShape_xy[1]-1],
              [fullImShape_xy[0]-1, 0]
              ])

def resize_pad(img: NChanImg, newSize: Tuple[int, int], interp=cv.INTER_NEAREST, padVal=0):
  """
  Resizes image to the requested size using the specified interpolation method.
  :param img: Image to resize
  :param newSize: New size for image. Since aspect ratio is maintained, the portion of the
    image which couldn't be resized fully is padded with a constant value of `padVal`.
    For instance, if the original image is 5x10 and the requested new size is 10x15, then
    after resizing the image will be 7x15 to preserve aspect ratio. 2 pixels of padding
    will be added on the left and 1 pixel of padding will be added on the right so the final
    output is 10x15.
  :param padVal: Value to pad dimension that couldn't be fully resized
  :param interp: Interpolation method to use during resizing
  :param padVal: Constant value to use during padding
  :return: Resized and padded image
  """
  needsRotate = False
  # Make sure largest dimension is along rows for easier treatment and rotate back after
  newSize = np.array(newSize)
  resizeRatio = min(newSize/np.array(img.shape[:2]))
  paddedImg = cv.resize(img, (0, 0), fx = resizeRatio, fy = resizeRatio, interpolation=interp)
  padDimension = np.argmax(newSize - np.array(paddedImg.shape[:2]))
  if padDimension != 1:
    paddedImg = cv.rotate(paddedImg, cv.ROTATE_90_CLOCKWISE)
    newSize = newSize[::-1]
    needsRotate = True
  # Now pad dimension is guaranteed to be index 1
  padDimension = 1
  padding = (newSize[padDimension] - paddedImg.shape[padDimension]) // 2
  if padding > 0:
    paddedImg = cv.copyMakeBorder(paddedImg, 0, 0, padding, padding, cv.BORDER_CONSTANT, value=padVal)
  # Happens during off-by-one truncated division
  remainingPadding = newSize[padDimension] - paddedImg.shape[padDimension]
  if remainingPadding > 0:
    paddedImg = cv.copyMakeBorder(paddedImg, 0, 0, remainingPadding, 0, cv.BORDER_CONSTANT, value=padVal)
  if needsRotate:
    paddedImg = cv.rotate(paddedImg, cv.ROTATE_90_COUNTERCLOCKWISE)
  return paddedImg

@contextmanager
def monkeyPatch(obj, toChange: str, newVal):
  oldVal = getattr(obj, toChange, None)
  setattr(obj, toChange, newVal)
  yield
  if oldVal is None:
    delattr(obj, toChange)
  else:
    setattr(obj, toChange, oldVal)

def hierarchicalUpdate(curDict: dict, other: dict):
  """Dictionary update that allows nested keys to be updated without deleting the non-updated keys"""
  if other is None:
    return curDict
  for k, v in other.items():
    curVal = curDict.get(k, None)
    if isinstance(curVal, dict) and isinstance(v, dict):
      hierarchicalUpdate(curVal, v)
    elif isinstance(curVal, list) and isinstance(v, list):
      curVal.extend(v)
    else:
      curDict[k] = v
  return curDict

def showMaskDiff(oldMask: BlackWhiteImg, newMask: BlackWhiteImg):
  infoMask = np.tile(oldMask[...,None].astype('uint8')*255, (1,1,3))
  # Was there, now it's not -- color red
  infoMask[oldMask & ~newMask,:] = [255,0,0]
  # Wasn't there, now it is -- color green
  infoMask[~oldMask & newMask,:] = [0,255,0]
  return infoMask

# Poor man's LRU dict, since I don't yet feel like including another pypi dependency
class MaxSizeDict(dict):
  def __init__(self, *args, maxsize:int=np.inf, **kwargs):
    super().__init__(*args, **kwargs)
    self.maxsize = maxsize

  def __setitem__(self, key, value):
    if len(self) == self.maxsize:
      # Evict oldest inserted entry
      self.pop(next(iter(self.keys())))
    super().__setitem__(key, value)


from functools import partial
from typing import Sequence
from warnings import warn

import numpy as np
import pandas as pd
from pandas import DataFrame as df
from pyqtgraph.Qt import QtWidgets, QtCore, QtGui

from s3a.constants import PRJ_CONSTS, REQD_TBL_FIELDS, PRJ_ENUMS
from s3a.models.tablemodel import ComponentMgr
from s3a.parameditors.singleton import PRJ_SINGLETON
from s3a.structures import TwoDArr

__all__ = ['CompTableView']

from utilitys import ParamEditor, EditorPropsMixin, RunOpts
from utilitys.params.pgregistered import PgParamDelegate

Signal = QtCore.Signal

TBL_FIELDS = PRJ_SINGLETON.tableData.allFields

class PopupTableDialog(QtWidgets.QDialog):
  def __init__(self, *args):
    super().__init__(*args)
    self.setModal(True)
    # -----------
    # Table View
    # -----------
    self.tbl = CompTableView(minimal=True)
    # Keeps track of which columns were edited by the user
    self.dirtyColIdxs = []

    # -----------
    # Warning Message
    # -----------
    self.reflectDelegateChange()
    self.warnLbl = QtWidgets.QLabel(self)
    self.warnLbl.setStyleSheet("font-weight: bold; color:red; font-size:14")
    self.updateWarnMsg([])

    # -----------
    # Widget buttons
    # -----------
    self.applyBtn = QtWidgets.QPushButton('Apply')
    self.closeBtn = QtWidgets.QPushButton('Close')

    # -----------
    # Widget layout
    # -----------
    btnLayout = QtWidgets.QHBoxLayout()
    btnLayout.addWidget(self.applyBtn)
    btnLayout.addWidget(self.closeBtn)

    centralLayout = QtWidgets.QVBoxLayout()
    centralLayout.addWidget(self.warnLbl)
    centralLayout.addWidget(self.tbl)
    centralLayout.addLayout(btnLayout)
    self.setLayout(centralLayout)
    self.setMinimumWidth(self.tbl.width())

    # -----------
    # UI Element Signals
    # -----------
    self.closeBtn.clicked.connect(self.close)
    self.applyBtn.clicked.connect(self.accept)

  def updateWarnMsg(self, updatableCols: Sequence[str]):
    warnMsg = 'Note! '
    if len(updatableCols) == 0:
      tblInfo = 'No information'
    else:
      tblInfo = "Only " + ", ".join(updatableCols)
    warnMsg += f'{tblInfo} will be updated from this view.'
    self.warnLbl.setText(warnMsg)

  def reflectDelegateChange(self):
    # TODO: Find if there's a better way to see if changes happen in a table
    self.titles = np.array(list([f.name for f in TBL_FIELDS]))
    self.tbl.setColDelegates()
    for colIdx in range(len(self.titles)):
      deleg = self.tbl.itemDelegateForColumn(colIdx)
      deleg.commitData.connect(partial(self.reflectDataChanged, colIdx))


  @property
  def data(self):
    return self.tbl.mgr.compDf.iloc[[0],:]

  def setData(self, compDf: df, colIdxs: Sequence):
    # New selection, so reset dirty columns
    self.dirtyColIdxs = []
    # Hide columns that weren't selected by the user since these changes
    # Won't propagate
    for ii in range(len(self.titles)):
      if ii not in colIdxs:
        self.tbl.hideColumn(ii)
      else:
        self.tbl.showColumn(ii)
    self.tbl.mgr.rmComps()
    self.tbl.mgr.addComps(compDf, addtype=PRJ_ENUMS.COMP_ADD_AS_MERGE)
    self.updateWarnMsg([])

  def reflectDataChanged(self, editedColIdx: int):
    """Responds to user editing the value in a cell by setting a dirty bit for that column"""
    if editedColIdx not in self.dirtyColIdxs:
      self.dirtyColIdxs.append(editedColIdx)
      self.dirtyColIdxs.sort()
      self.updateWarnMsg(self.titles[self.dirtyColIdxs])

  def reject(self):
    # On dialog close be sure to unhide all columns / reset dirty cols
    self.dirtyColIdxs = []
    for ii in range(len(self.titles)):
      self.tbl.showColumn(ii)
    super().reject()

class CompTableView(EditorPropsMixin, QtWidgets.QTableView):
  __groupingName__ = 'Component Table'
  """
  Table for displaying :class:`ComponentMgr` data.
  """
  sigSelectionChanged = Signal(object)

  @classmethod
  def __initEditorParams__(cls):
    cls.showOnCreate = PRJ_SINGLETON.generalProps.registerProp(PRJ_CONSTS.PROP_SHOW_TBL_ON_COMP_CREATE)
    cls.toolsEditor = ParamEditor.buildClsToolsEditor(cls, name='Component Table Tools')

  def __init__(self, *args, minimal=False):
    """
    Creates the table.

    :param minimal: Whether to make a table with minimal features.
       If false, creates extensible table with context menu options.
       Otherwise, only contains minimal features.
    """
    super().__init__(*args)

    self.setStyleSheet("QTableView { selection-color: white; selection-background-color: #0078d7; }")
    self._prevSelRows = np.array([])
    self.setSortingEnabled(True)

    self.mgr = ComponentMgr()
    self.minimal = minimal
    self.setModel(self.mgr)
    self.setColDelegates()

    with PRJ_SINGLETON.generalProps.setBaseRegisterPath(self.__groupingName__):
      proc, params = PRJ_SINGLETON.generalProps.registerFunc(
        self.setVisibleColumns, runOpts=RunOpts.ON_CHANGED, nest=False,
        returnParam=True, visibleColumns=[])
    def onChange(*_args):
      params.child('visibleColumns').setLimits([f.name for f in TBL_FIELDS])
    onChange()
    PRJ_SINGLETON.tableData.sigCfgUpdated.connect(onChange)
    proc.run()

    if not minimal:
      self.popup = PopupTableDialog(*args)
      # Create context menu for changing table rows
      self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
      cursor = QtGui.QCursor()
      self.customContextMenuRequested.connect(lambda: self.menu.exec_(cursor.pos()))

    self.instIdColIdx = TBL_FIELDS.index(REQD_TBL_FIELDS.INST_ID)

  def setVisibleColumns(self, visibleColumns: Sequence[str]):
    """
    Determines which columns to show. All unspecified columns will be hidden.
    :param visibleColumns:
      pType: checklist
    """
    for ii, col in enumerate(self.mgr.colTitles):
      self.setColumnHidden(ii, col not in visibleColumns)

  def setColDelegates(self):
    for ii, field in enumerate(TBL_FIELDS):
      curType = field.pType.lower()
      curval = field.value
      paramDict = dict(type=curType, default=curval, **field.opts)
      if curType == 'enum':
        paramDict['type'] = 'list'
        paramDict.update(values=list(type(curval)))
      elif curType == 'prjparam':
        paramDict['type'] = 'list'
        paramDict.update(values=list(curval.group))
      elif curType == 'bool':
        # TODO: Get checkbox to stay in table after editing for a smoother appearance.
        #   For now, the easiest solution is to just use dropdown
        paramDict['type'] = 'list'
        paramDict.update(values={'True': True, 'False': False})
      try:
        self.setItemDelegateForColumn(ii, PgParamDelegate(paramDict, self))
      except TypeError:
        # Parameter doesn't have a registered pyqtgraph editor, so default to
        # generic text editor
        paramDict['type'] = 'text'
        paramDict['default'] = str(curval)
        self.setItemDelegateForColumn(ii, PgParamDelegate(paramDict, self))

    self.horizontalHeader().setSectionsMovable(True)

  # When the model is changed, get a reference to the ComponentMgr
  def setModel(self, modelOrProxy: QtCore.QAbstractTableModel):
    super().setModel(modelOrProxy)
    try:
      # If successful we were given a proxy model
      self.mgr = modelOrProxy.sourceModel()
    except AttributeError:
      self.mgr = modelOrProxy

  def selectionChanged(self, curSel: QtCore.QItemSelection, prevSel: QtCore.QItemSelection):
    """
    When the selected rows in the table change, this retrieves the corresponding previously
    and newly selected IDs. They are then emitted in sigSelectionChanged.
    """
    super().selectionChanged(curSel, prevSel)
    selectedIds = []
    selection = self.selectionModel().selectedIndexes()
    for item in selection:
      selectedIds.append(item.sibling(item.row(),self.instIdColIdx).data(QtCore.Qt.EditRole))
    newRows = pd.unique(selectedIds)
    if np.array_equal(newRows, self._prevSelRows):
      return
    self._prevSelRows = newRows
    self.sigSelectionChanged.emit(pd.unique(selectedIds))

  def removeSelectedRows_gui(self):
    if self.minimal: return

    idList = [idx.siblingAtColumn(self.instIdColIdx).data(QtCore.Qt.EditRole)
              for idx in self.selectedIndexes()]
    if len(idList) == 0:
      return
    # Make sure the user actually wants this
    dlg = QtWidgets.QMessageBox()
    confirm  = dlg.question(self, 'Remove Rows', 'Are you sure you want to remove these rows?',
                 dlg.Yes | dlg.Cancel)
    if confirm == dlg.Yes:
      # Proceed with operation
      # Since each selection represents a row, remove duplicate row indices
      idList = pd.unique(idList)
      self.mgr.rmComps(idList)
      self.clearSelection()

  def ids_rows_colsFromSelection(self, excludeNoEditCols=True, warnNoneSelection=True):
    """Returns Nx3 np array of (ids, rows, cols) from current table selection"""
    selectedIdxs = self.selectedIndexes()
    retLists = [] # (Ids, rows, cols)
    for idx in selectedIdxs:
      row = idx.row()
      # 0th row contains instance ID
      # TODO: If the user is allowed to reorder columns this needs to be revisited
      idAtIdx = idx.siblingAtColumn(self.instIdColIdx).data(QtCore.Qt.EditRole)
      retLists.append([idAtIdx, row, idx.column()])
    retLists = np.array(retLists)
    if excludeNoEditCols and len(retLists) > 0:
      # Set diff will eliminate any repeats, so use a slower op that at least preserves
      # duplicates
      retLists = retLists[~np.isin(retLists[:,2], self.mgr.noEditColIdxs)]
    if warnNoneSelection and len(retLists) == 0:
      warn('No editable columns selected.', UserWarning)
    return retLists

  def setSelectedCellsAsFirst(self):
    """
    Sets all cells in the selection to be the same as the first row in the selection.
    See the project wiki for a detailed description
    """
    selection = self.ids_rows_colsFromSelection()
    overwriteData = self.mgr.compDf.loc[selection[0,0]]
    self.setSelectedCellsAs(selection, overwriteData)

  def setSelectedCellsAs_gui(self, selectionIdxs: TwoDArr=None):
    """
    Sets all cells in the selection to the values specified in the popup table. See
    the project wiki for a detailed description
    """
    if selectionIdxs is None:
      selectionIdxs = self.ids_rows_colsFromSelection()
    overwriteData = self.mgr.compDf.loc[[selectionIdxs[0,0]]].copy()
    with PRJ_SINGLETON.actionStack.ignoreActions():
      self.popup.setData(overwriteData, pd.unique(selectionIdxs[:,2]))
      wasAccepted = self.popup.exec_()
    if not wasAccepted or len(self.popup.dirtyColIdxs) == 0:
      return

    selectionIdxs = selectionIdxs[np.isin(selectionIdxs[:,2], self.popup.dirtyColIdxs)]
    self.setSelectedCellsAs(selectionIdxs, self.popup.data)

  def setSelectedCellsAs(self, selectionIdxs: TwoDArr, overwriteData: df):
    """
    Overwrites the data from rows and cols with the information in *overwriteData*.
    Each (id, row, col) index is treated as a single index
    :param selectionIdxs: Selection idxs to overwrite. If *None*, defaults to
      current selection.
    :param overwriteData: What to fill in the overwrite locations. If *None*, a popup table
      is displayed and its data is used.
    """
    if self.minimal: return

    if len(selectionIdxs) == 0:
      return
    overwriteData = overwriteData.squeeze()
    uniqueIds = pd.unique(selectionIdxs[:,0])
    newDataDf = self.mgr.compDf.loc[uniqueIds].copy()
    # New data ilocs will no longer match, fix this using loc + indexed columns
    colsForLoc = self.mgr.compDf.columns[selectionIdxs[:,2]]
    for idxTriplet, colForLoc in zip(selectionIdxs, colsForLoc):
      newDataDf.at[idxTriplet[0], colForLoc] = overwriteData.iat[idxTriplet[2]]
    self.mgr.addComps(newDataDf, addtype=PRJ_ENUMS.COMP_ADD_AS_MERGE)


import numpy as np
import pyqtgraph as pg
from pyqtgraph.GraphicsScene.mouseEvents import MouseDragEvent
from pyqtgraph.Qt import QtCore, QtGui
from skimage.measure import points_in_poly

from s3a.structures import XYVertices

Signal = QtCore.Signal

__all__ = ['BoundScatterPlot', 'RightPanViewBox']

class BoundScatterPlot(pg.ScatterPlotItem):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    # TODO: Find out where the mouse is and make sure it's above a point before changing
    # the mouse cursor

    self.hoverCursor = QtCore.Qt.PointingHandCursor

  # Not working at the moment :/
  # def mouseMoveEvent(self, ev):
  #   if self.pointsAt(ev.pos()):
  #     self.setCursor(self.hoverCursor)
  #   else:
  #     self.unsetCursor()

  def boundsWithin(self, selection: XYVertices):
    # TODO: Optimize for rectangular selections
    # polyPoints = [QtCore.QPointF(*row) for row in selection]
    # selectionPoly = QtGui.QPolygonF(polyPoints)
    if self.data['x'] is None:
      return np.array([])
    pointLocs = np.column_stack(self.getData())
    # tfIsInSelection = (pointLocs[0] >= bbox[0]) \
    #   & (pointLocs[0] <= bbox[2]) \
    #   & (pointLocs[1] >= bbox[1]) \
    #   & (pointLocs[1] <= bbox[3])
    tfIsInSelection = points_in_poly(pointLocs, selection)
    # tfIsInSelection = np.array([selectionPoly.containsPoint(QtCore.QPointF(*row), QtCore.Qt.WindingFill)
    #                             for row in pointLocs], dtype=bool)
    return np.array([point.data() for point in self.points()[tfIsInSelection]])

  def pointsAt(self, pos: QtCore.QPointF, boundaryOnly=True):
    """
    The default implementation only checks a square around each spot. However, this is not
    precise enough for my needs. The overloaded pointsAt checks any polygonal area as defined
    by the spot boundary. It also triggers when clicking *inside* the spot boundary,
    which I don't want.

    :param pos: Where to search for points
    :param boundaryOnly: If `True`, only locates points whose boundary is on the pos.
      Otherwise, any points containing pos will be returned.
    """
    pts = []
    spots = self.points()
    if len(spots) == 0:
      return spots
    strokerWidth = spots[0].pen().width()
    for spot in spots: # type: pg.SpotItem
      symb = QtGui.QPainterPath(spot.symbol())
      symb.translate(spot.pos())
      stroker = QtGui.QPainterPathStroker()
      stroker.setWidth(strokerWidth)
      checkSymb = symb
      if boundaryOnly:
        mousePath = stroker.createStroke(symb)
        checkSymb = mousePath
      # Only trigger when clicking a boundary, not the inside of the shape
      if checkSymb.contains(pos):
        pts.append(spot)
    return pts[::-1]

  def _updateMaxSpotSizes(self, **kwargs):
    w = -1
    try:
      for symbol in self.data['symbol']:
        br = symbol.boundingRect()
        w = max(w, br.width()*2, br.height()*2)
    except AttributeError:
      # Don't have a painter symbol
      return super()._updateMaxSpotSizes(**kwargs)
    self._maxSpotWidth = max(self._maxSpotPxWidth, w)
    self._maxSpotPxWidth = w
    self.bounds = [None, None]

class RightPanViewBox(pg.ViewBox):
  def mouseDragEvent(self, ev: MouseDragEvent, axis=None):
    if ev.buttons() == QtCore.Qt.RightButton \
        or ev.button() == QtCore.Qt.RightButton:
      ev.buttons = lambda: QtCore.Qt.LeftButton
      ev.button = ev.buttons
    super().mouseDragEvent(ev)


from dataclasses import dataclass
from enum import Enum
from pathlib import Path

__all__ = ['BASE_DIR', 'MENU_OPTS_DIR', 'ICON_DIR', 'QUICK_LOAD_DIR',
           'SCHEMES_DIR', 'LAYOUTS_DIR', 'TABLE_DIR', 'GEN_PROPS_DIR', 'SHORTCUTS_DIR',
           'SHORTCUT_BASE', 'MAIN_IMG_DIR', 'APP_STATE_DIR',
           'REQD_TBL_FIELDS', 'PRJ_CONSTS', 'PRJ_ENUMS', 'PROJ_FILE_TYPE', 'PROJ_BASE_TEMPLATE']

from typing import Any

BASE_DIR = Path(__file__).parent
MENU_OPTS_DIR = BASE_DIR/'menuopts'
ICON_DIR = BASE_DIR/'icons'

# -----
# EDITORS
# -----
LAYOUTS_DIR = MENU_OPTS_DIR/'layouts'
QUICK_LOAD_DIR = MENU_OPTS_DIR/'quickloaders'
SCHEMES_DIR = MENU_OPTS_DIR/'schemes'
GEN_PROPS_DIR = MENU_OPTS_DIR/'genprops'
SHORTCUTS_DIR = MENU_OPTS_DIR/'shortcuts'
MAIN_IMG_DIR = MENU_OPTS_DIR/'mainimg'
TABLE_DIR = MENU_OPTS_DIR/'table'
APP_STATE_DIR = Path.home()/'.s3a'


PROJ_BASE_TEMPLATE = BASE_DIR/'projectcfg.yml'
PROJ_FILE_TYPE = 's3aprj'

# Ensure menuopts and layouts directories exist
LAYOUTS_DIR.mkdir(parents=True, exist_ok=True)
APP_STATE_DIR.mkdir(parents=True, exist_ok=True)

SHORTCUT_BASE = 'Ctrl+Q'

class _PrjEnums(Enum):
  # --------------------------
  # COMPONENTS
  # --------------------------
  COMP_ADD_AS_MERGE = 'Add as Merge'
  COMP_ADD_AS_NEW   = 'Add as New'
  COMP_EXPORT_ALL   = 'Export All Components'
  COMP_RM_ALL       = "Remove All"

  # --------------------------
  # REGION CREATION
  # --------------------------
  BUFFER_UNDO       = 'Undo'
  BUFFER_REDO       = 'Redo'

  # --------------------------
  # VERTICES
  # --------------------------
  HIER_ALL_FILLED  = 'All Filled'

  # --------------------------
  # COMMON DATAFRAME FIELDS
  # --------------------------
  FIELD_SELECTED   = 'selected'
  FIELD_FOCUSED    = 'focused'
  FIELD_LABEL    = 'label'

PRJ_ENUMS = _PrjEnums

from s3a.structures import ComplexXYVertices, PrjParam, PrjParamGroup, newParam

@dataclass
class _ReqdTableFields(PrjParamGroup):
  _extraRequired = []

  INST_ID          : PrjParam = newParam('Instance ID', -1, readonly=True)
  VERTICES         : PrjParam = newParam('Vertices', ComplexXYVertices(), readonly=True)
  SRC_IMG_FILENAME : PrjParam = newParam('Source Image Filename', "", readonly=True)

  def addField(self, field: PrjParam):
    if field not in self:
      self._extraRequired.append(field)

  def removeField(self, field: PrjParam):
    if field in self._extraRequired:
      self._extraRequired.pop(field)

  def __iter__(self):
    yield from super().__iter__()
    for field in self._extraRequired:
      yield field


REQD_TBL_FIELDS = _ReqdTableFields()


@dataclass
class _PrjConsts(PrjParamGroup):
  # --------------------------
  # CLASS NAMES
  # --------------------------
  CLS_ANNOTATOR        : PrjParam = newParam('S3A Window')
  CLS_S3A_MODEL        : PrjParam = newParam('S3A Model')

  CLS_COMP_TBL         : PrjParam = newParam('Component Table')
  CLS_COMP_MGR         : PrjParam = newParam('Component Manager')
  CLS_COMP_EXPORTER    : PrjParam = newParam('Component Exporter')

  CLS_VERT_IMG         : PrjParam = newParam('Focused Image Graphics')
  CLS_MULT_REG_PLT     : PrjParam = newParam('Component Plot')
  CLS_ROI_CLCTN        : PrjParam = newParam('ROI Shapes')

  CLS_REGION_BUF       : PrjParam = newParam('Region Modification Buffer')

  CLS_IMG_AREA         : PrjParam = newParam('Base Image Area')
  CLS_MAIN_IMG_AREA    : PrjParam = newParam('Main Image Area')
  CLS_FOCUSED_IMG_AREA : PrjParam = newParam('Focused Image Area')
  # --------------------------
  # SCHEME PARAMETERS
  # --------------------------
  SCHEME_USE_DARK_THEME  : PrjParam = newParam('Use dark theme', False)
  SCHEME_BOUND_CLR       : PrjParam = newParam('Normal Boundary Color', 'ff0', 'color')
  SCHEME_SEL_BOUND_CLR   : PrjParam = newParam('Selected Boundary Color', '00f', 'color')
  SCHEME_FOC_BRUSH_CLR   : PrjParam = newParam('Focused Brush Color', 'f007', 'color')
  SCHEME_BOUND_WIDTH     : PrjParam = newParam('Boundary Width', 7)
  SCHEME_REG_VERT_COLOR  : PrjParam = newParam('Vertex Color', '0f0', 'color')
  SCHEME_REG_FILL_COLOR  : PrjParam = newParam('Fill Color', '00ff0046', 'color')
  SCHEME_ROI_LINE_CLR    : PrjParam = newParam('ROI Line Color', 'fff', 'color')
  SCHEME_ROI_LINE_WIDTH  : PrjParam = newParam('ROI Line Width', 1)
  SCHEME_GRID_LINE_WIDTH : PrjParam = newParam('Grid Line Width', 1)
  SCHEME_GRID_CLR        : PrjParam = newParam('Grid Line Color', 'fff', 'color')
  SCHEME_SHOW_GRID       : PrjParam = newParam('Show Grid', False)
  SCHEME_LBL_COL         : PrjParam = newParam('Labeling Column', REQD_TBL_FIELDS.INST_ID.name, pType='list', limits=[])

  # --------------------------
  # REGION-CREATION PARAMETERS
  # --------------------------
  PROP_MIN_COMP_SZ             : PrjParam = newParam('Minimum New Component Size (px)', 50)
  PROP_UNDO_BUF_SZ             : PrjParam = newParam('Undo Buffer Size', 300)
  PROP_COMP_SEL_BHV            : PrjParam = newParam(
    'Component Selection Behavior', 'Entire Component', 'list', limits=['Boundary Only', 'Entire Component'],
    helpText='When `Boundary Only`, components are selected when clicking on their boundaries.\n'
             'When `Entire Component`, clicking anywhere within the component'
             ' will select it')
  PROP_SHOW_TBL_ON_COMP_CREATE : PrjParam = newParam('Show popup table when creating component', False)
  # --------------------------
  # MISC TOOLS
  # --------------------------
  TOOL_UNDO               : PrjParam =  newParam('Undo', 'Ctrl+Z')
  TOOL_REDO               : PrjParam =  newParam('Redo', 'Ctrl+Y')
  # --------------------------
  # IMAGE TOOLS
  # --------------------------
  TOOL_MERGE_COMPS        : PrjParam = newParam('Merge Selected', f'{SHORTCUT_BASE},S,M', icon=str(ICON_DIR/'merge.svg'))
  TOOL_SPLIT_COMPS        : PrjParam = newParam('Split Selected', f'{SHORTCUT_BASE},S,S', icon=str(ICON_DIR/'split.svg'))
  TOOL_REM_OVERLAP        : PrjParam = newParam('Remove Component Overlap', f'{SHORTCUT_BASE},S,E', icon=str(ICON_DIR/'mutex.svg'))
  TOOL_COPY_REGIONS       : PrjParam = newParam('Copy Selected', f'{SHORTCUT_BASE},S,C', icon=str(ICON_DIR/'copy.svg'))
  TOOL_MOVE_REGIONS       : PrjParam = newParam('Move Selected', f'{SHORTCUT_BASE},S,V', icon=str(ICON_DIR/'move.svg'))
  TOOL_CLEAR_FOC_REGION   : PrjParam = newParam('Clear', f'{SHORTCUT_BASE},V,C', icon=str(ICON_DIR/'clear.svg'))
  TOOL_RESET_FOC_REGION   : PrjParam = newParam('Reset', f'{SHORTCUT_BASE},V,R', icon=str(ICON_DIR/'reset.svg'))
  TOOL_FILL_FOC_REGION    : PrjParam = newParam('Fill', f'{SHORTCUT_BASE},V,F', icon=str(ICON_DIR/'fill.svg'))
  TOOL_ACCEPT_FOC_REGION  : PrjParam = newParam('Accept', 'Ctrl+Shift+A', icon=str(ICON_DIR/'accept.svg'))
  TOOL_CLEAR_ROI          : PrjParam = newParam('Clear ROI', 'Esc')
  TOOL_CLEAR_HISTORY      : PrjParam = newParam('Clear Processor History', f'{SHORTCUT_BASE},V,H', icon=str(ICON_DIR/'clear_history.svg'))
  TOOL_PROC_ANALYTICS     : PrjParam = newParam('Show Analytics', icon=str(ICON_DIR/'analytics.svg'))
  TOOL_RESET_ZOOM         : PrjParam = newParam('Reset Zoom', f'{SHORTCUT_BASE},Z,R', icon=str(ICON_DIR/'reset_zoom.svg'))

  # --------------------------
  # WINDOW TOOLS
  # --------------------------
  TOOL_ESTIMATE_BOUNDARIES   : PrjParam = newParam('Estimate Boundaries', 'Ctrl+Alt+Shift+E')
  TOOL_CLEAR_BOUNDARIES      : PrjParam = newParam('Clear Boundaries', 'Ctrl+Alt+Shift+C')
  TOOL_EXPORT_COMP_LIST      : PrjParam = newParam('Export Current Table', f'{SHORTCUT_BASE},E,T')
  TOOL_TBL_SET_SAME_AS_FIRST : PrjParam = newParam('Set Cells as First', 'Ctrl+D')
  TOOL_TBL_SET_AS            : PrjParam = newParam('Set Cells As...', 'Ctrl+Shift+D')
  TOOL_TBL_DEL_ROWS          : PrjParam = newParam('Delete Table Rows', 'Del')
  TOOL_TBL_ZOOM_TO_COMPS     : PrjParam = newParam('Zoom to Selection  ', f'{SHORTCUT_BASE},Z,S')

  # --------------------------
  # PROJECT
  # --------------------------
  TOOL_PROJ_SAVE     : PrjParam = newParam('Save', 'Ctrl+S')
  TOOL_PROJ_OPEN     : PrjParam = newParam('Open Project', f'{SHORTCUT_BASE},P,O')
  TOOL_PROJ_OPEN_IMG : PrjParam = newParam('Open Project Image', f'{SHORTCUT_BASE},I,O')
  TOOL_PROJ_CREATE   : PrjParam = newParam('Create Project', f'{SHORTCUT_BASE},P,C')
  TOOL_PROJ_ADD_IMG  : PrjParam = newParam('Add New Image', f'{SHORTCUT_BASE},I,A')
  TOOL_PROJ_ADD_ANN  : PrjParam = newParam('Add New Annotation', f'{SHORTCUT_BASE},A,A')
  TOOL_PROJ_SETTINGS : PrjParam = newParam('Project Settings...', f'{SHORTCUT_BASE},P,S')
  TOOL_PROJ_EXPORT   : PrjParam = newParam('Export...', f'{SHORTCUT_BASE},P,E')
  TOOL_AUTOSAVE      : PrjParam = newParam('Autosave...', f'{SHORTCUT_BASE},A,O')

  # --------------------------
  # GLOBAL PREDICTIONS
  # --------------------------
  TOOL_PRED_SEL     : PrjParam = newParam('Predict From Selection', f'{SHORTCUT_BASE},M,P', icon=str(ICON_DIR/'predict.svg'))


  # --------------------------
  # COMPONENT EXPORT PARAMETERS
  # --------------------------
  EXP_ONLY_VISIBLE   : PrjParam = newParam(
    'Only Export Visible Components', False,
    helpText='If *True*, only components showing on the main image will be included in'
             ' file exports.')
  INCLUDE_FNAME_PATH : PrjParam = newParam('Include full image path on export', False,
  None, 'If *False*, only the image name is included in export information, e.g. <img>.png.'
        ' Otherwise, the full filepath is included, e.g. "C:/Users/.../<img>.png')

  # --------------------------
  # COMPONENT EXPORT PARAMETERS
  # --------------------------
  ANN_CUR_FILE_INDICATOR: PrjParam = newParam('Current image file string', 'Newly Added')

  # --------------------------
  # DRAWING
  # -------------------
  # Modes
  DRAW_MODE_FOCUSED : PrjParam = newParam(
    'Activate "Edit" draw mode', f'{SHORTCUT_BASE},D,E', 'registeredaction',
    icon=str(ICON_DIR/'edit.svg'))

  # Shapes
  DRAW_SHAPE_RECT : PrjParam = newParam(
    'Activate "Rectangular" draw shape', f'{SHORTCUT_BASE},D,R', 'registeredaction',
    icon=str(ICON_DIR/'rectangle.svg'))
  DRAW_SHAPE_POLY : PrjParam = newParam(
    'Activate "Polygon" draw shape', f'{SHORTCUT_BASE},D,Y', 'registeredaction',
    icon=str(ICON_DIR/'polygon.svg'))
  DRAW_SHAPE_ELLIPSE : PrjParam = newParam(
    'Activate "Ellipse" draw shape', f'{SHORTCUT_BASE},D,L', 'registeredaction',
    icon=str(ICON_DIR/'ellipse.svg'))
  DRAW_SHAPE_FREE : PrjParam = newParam('Activate "Freehand" draw shape', f'{SHORTCUT_BASE},D,H',
                                        icon=str(ICON_DIR/'freehand.svg'))
  DRAW_SHAPE_POINT: PrjParam = newParam(
    'Activate "Point" draw shape', f'{SHORTCUT_BASE},D,N', icon=str(ICON_DIR / 'point.svg'))
  DRAW_SHAPE_NONE : PrjParam = newParam('None')

  # Actions
  DRAW_ACT_CREATE    : PrjParam = newParam(
    'Activate "Create Component" action', f'{SHORTCUT_BASE},D,C', 'registeredaction',
    icon=str(ICON_DIR/'create.svg'),
    helpText='When an ROI is created, the image processor will attempt to make a new'
             ' component at that location. Right-click and drag to pan.')
  DRAW_ACT_ADD    : PrjParam = newParam(
    'Activate "Add to Foreground" action', f'{SHORTCUT_BASE},D,F', 'registeredaction',
    icon=str(ICON_DIR/'foreground.svg'),
    helpText='When an ROI is created, the image processor will attempt to make a new'
             ' component at that location. Right-click and drag to pan.')
  DRAW_ACT_REM    : PrjParam = newParam(
    'Activate "Add to Background" action', f'{SHORTCUT_BASE},D, B', 'registeredaction',
    icon=str(ICON_DIR/'background.svg'),
    helpText='When an ROI is created, the image processor will attempt to take the enclosed'
             ' area away from the current component shape. Right-click and drag to pan.')
  DRAW_ACT_SELECT : PrjParam = newParam(
    'Activate "Select" draw action', f'{SHORTCUT_BASE},D, S', 'registeredaction', icon=str(ICON_DIR / 'select.svg'),
    helpText='When component boundaries are enclosed by this ROI, they will be selected'
             ' in the component table. Right-click and drag to pan.')
  DRAW_ACT_PAN    : PrjParam = newParam(
    'Activate "Pan" draw action', f'{SHORTCUT_BASE},D,P', 'registeredaction', icon=str(ICON_DIR / 'pan.svg'),
    helpText='No ROI will be drawn in this mode. Right- or left-click and drag to pan.')
PRJ_CONSTS = _PrjConsts()

# from ruamel.yaml import YAML
# yaml = YAML()
# for cls in PrjParam, PrjParamGroup, _PrjConsts:
#   yaml.register_class(cls)
# for p in PRJ_CONSTS:
#   p.group = []
# p = Path('./consts.yml')
# yaml.dump(PRJ_CONSTS, p)



from pathlib import Path

from pyqtgraph.Qt import QtGui, QtWidgets, QtCore
import pyqtgraph as pg
import fire

def _s3agui(**kwargs):
  app = pg.mkQApp()
  here = Path(__file__).parent
  splash = QtWidgets.QSplashScreen(QtGui.QPixmap(str(here/'icons'/'s3asplash.svg')),
                                   QtCore.Qt.WindowStaysOnTopHint)
  splash.show()
  app.processEvents()
  from s3a import S3A, appInst

  win = S3A(**kwargs)
  splash.finish(win)

  QtCore.QTimer.singleShot(0, win.showMaximized)
  appInst.exec_()

def s3agui():
  fire.Fire(_s3agui)

if __name__ == '__main__':
    s3agui()


from __future__ import annotations

from typing import List, Type, Dict

from pyqtgraph.Qt import QtWidgets, QtCore
from utilitys import ActionStack
from utilitys import ParamEditor, ParamEditorDockGrouping, NestedProcWrapper, ParamEditorPlugin, dockPluginFactory

from s3a.constants import GEN_PROPS_DIR, SCHEMES_DIR, SHORTCUTS_DIR
from .algcollection import AlgCtorCollection
from .quickloader import QuickLoaderEditor
from .table import TableData
from ..processing import ImgProcWrapper
from utilitys.params import ShortcutParameter

Signal = QtCore.Signal


class _PrjSingleton(QtCore.QObject):
  sigPluginAdded = Signal(object) # List[QtWidgets.QDockWidget]

  def __init__(self, parent=None):
    super().__init__(parent)
    self.actionStack = ActionStack()
    self.clsToPluginMapping: Dict[Type[ParamEditorPlugin], ParamEditorPlugin] = {}

    self.tableData = TableData()
    self.filter = self.tableData.filter


    self.generalProps = ParamEditor(saveDir=GEN_PROPS_DIR, fileType='genprops',
                                    name='App Settings')
    self.colorScheme = ParamEditor(saveDir=SCHEMES_DIR, fileType='scheme',
                                   name='Color Scheme')
    self.shortcuts = ShortcutParameter.setRegistry(createIfNone=True, saveDir=SHORTCUTS_DIR)
    self.quickLoader = QuickLoaderEditor()
    self.imgProcClctn = AlgCtorCollection(ImgProcWrapper)
    self.multiPredClctn = AlgCtorCollection(NestedProcWrapper)

    self.docks: List[QtWidgets.QDockWidget] = []
    self.addPlugin(dockPluginFactory('&Settings', [self.generalProps, self.colorScheme]))
    self.addPlugin(dockPluginFactory('Sho&rtcuts', [self.shortcuts, self.quickLoader]))

  @property
  def registerableEditors(self):
    outList = []
    for editor in self.docks:
      if isinstance(editor, ParamEditorDockGrouping):
        outList.extend(editor.editors)
      else:
        outList.append(editor)
    return outList

  def addPlugin(self, pluginCls: Type[ParamEditorPlugin], *args, **kwargs):
    """
    From a class inheriting the *PrjParamEditorPlugin*, creates a plugin object
    that will appear in the S3A toolbar. An entry is created with dropdown options
    for each editor in *pluginCls*'s *editors* attribute.

    :param pluginCls: Class containing plugin actions
    :param args: Passed to class constructor
    :param kwargs: Passed to class constructor
    """
    if pluginCls.name is not None and pluginCls.__groupingName__ is None:
      # The '&' character can be used in Qt to signal a shortcut, which is used internally
      # by plugins. When these plugins register shortcuts, the group name will contain
      # ampersands which shouldn't show up in human readable menus
      pluginCls.__groupingName__ = pluginCls.name.replace('&', '')
    plugin: ParamEditorPlugin = pluginCls(*args, **kwargs)
    self.clsToPluginMapping[pluginCls] = plugin
    self.sigPluginAdded.emit(plugin)
    if plugin.dock is not None and plugin.dock not in self.docks:
      self.docks.append(plugin.dock)
    return plugin

  def close(self):
    for editor in self.registerableEditors:
      editor.close()

INITIALIZED_GROUPINGS = set()
REGISTERED_GROUPINGS = set()
PRJ_SINGLETON = _PrjSingleton()


from typing import Dict, Collection

import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui
from utilitys import EditorPropsMixin, RunOpts

from s3a.parameditors import singleton
from s3a.structures import PrjParam, XYVertices
from s3a.views.rois import SHAPE_ROI_MAPPING, PlotDataROI, PointROI

__all__ = ['RoiCollection']

class RoiCollection(EditorPropsMixin, QtCore.QObject):
  # Signal(ExtendedROI)
  sigShapeFinished = QtCore.Signal(object) # roiVerts : XYVertices

  def __init__(self, allowableShapes: Collection[PrjParam]=(), parent: pg.GraphicsView=None):
    super().__init__(parent)
    singleton.PRJ_SINGLETON.colorScheme.registerFunc(
      PointROI.updateRadius, name='Point ROI Features', runOpts=RunOpts.ON_CHANGED, namePath=(self.__groupingName__,),

    )
    if allowableShapes is None:
      allowableShapes = set()
    self.shapeVerts = XYVertices()
    # Make a new graphics item for each roi type
    self.roiForShape: Dict[PrjParam, PlotDataROI] = {}
    self._curShape = next(iter(allowableShapes)) if len(allowableShapes) > 0 else None

    self._locks = set()
    self.addLock(self)
    self._parent = parent

    for shape in allowableShapes:
      newRoi = SHAPE_ROI_MAPPING[shape]()
      newRoi.setZValue(1000)
      self.roiForShape[shape] = newRoi
      newRoi.setRoiPoints()
      newRoi.hide()
    self.addRoisToView(parent)

  def addRoisToView(self, view: pg.GraphicsView):
    self._parent = view
    if view is not None:
      for roi in self.roiForShape.values():
        roi.hide()
        view.addItem(roi)

  def clearAllRois(self):
    for roi in self.roiForShape.values(): # type: PlotDataROI
      roi.setRoiPoints()
      roi.hide()
      self.addLock(self)

  def addLock(self, lock):
    """
    Allows this shape collection to be `locked`, preventing shapes from being drawn.
    Multiple locks can be applied; ROIs can only be drawn when all locks are removed.

    :param lock: Anything used as a lock. This will have to be manually removed later
      using `RoiCollection.removeLock`
    """
    self._locks.add(lock)

  def removeLock(self, lock):
    try:
      self._locks.remove(lock)
    except KeyError:
      pass

  def forceUnlock(self):
    self._locks.clear()

  @property
  def locked(self):
    return len(self._locks) > 0

  def buildRoi(self, ev: QtGui.QMouseEvent, imgItem: pg.ImageItem=None):
    """
    Construct the current shape ROI depending on mouse movement and current shape parameters
    :param imgItem: Image the ROI is drawn upon, used for mapping event coordinates
      from a scene to pixel coordinates. If *None*, event coordinates are assumed
      to already be relative to pixel coordinates.
    :param ev: Mouse event
    """
    # Unblock on mouse press
    # None imgitem is only the case during programmatic calls so allow this case
    if ((imgItem is None or imgItem.image is not None)
        and ev.type() == ev.MouseButtonPress
        and ev.button() == QtCore.Qt.LeftButton):
      self.removeLock(self)
    if self.locked: return False
    if imgItem is not None:
      posRelToImg = imgItem.mapFromScene(ev.pos())
    else:
      posRelToImg = ev.pos()
    # Form of rate-limiting -- only simulate click if the next pixel is at least one away
    # from the previous pixel location
    xyCoord = XYVertices([[posRelToImg.x(), posRelToImg.y()]], dtype=float)
    curRoi = self.curShape
    constructingRoi, self.shapeVerts = self.curShape.updateShape(ev, xyCoord)
    if self.shapeVerts is not None:
      self.sigShapeFinished.emit(self.shapeVerts)

    if not constructingRoi:
      # Vertices from the completed shape are already stored, so clean up the shapes.
      curRoi.setRoiPoints()
      curRoi.hide()
    else:
      # Still constructing ROI. Show it
      curRoi.show()
    return constructingRoi

  @property
  def curShapeParam(self): return self._curShape
  @curShapeParam.setter
  def curShapeParam(self, newShape: PrjParam):
    """
    When the shape is changed, be sure to reset the underlying ROIs
    :param newShape: New shape
    :return: None
    """
    # Reset the underlying ROIs for a different shape than we currently are using
    if newShape != self._curShape:
      self.clearAllRois()
    self._curShape = newShape

  @property
  def curShape(self):
      return self.roiForShape[self._curShape]